# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import sys
from contextlib import suppress as does_not_raise

import pytest
from flexmock import flexmock

import packit
from packit.actions import ActionName
from packit.actions_handler import ActionsHandler
from packit.exceptions import PackitException
from packit.upstream import Archive, GitUpstream, SRPMBuilder


@pytest.fixture
def upstream_pr_mock():
    return flexmock(url="test_pr_url")


@pytest.fixture
def tar_mock():
    def tar_mock_factory(archive_items=None, is_tarfile=True):
        if archive_items is None:
            archive_items = []
        tarfile_mock = flexmock(packit.upstream.tarfile)
        tarinfo_mock_list = [
            flexmock(  # noqa: PIE804
                **{
                    "name": name,
                    "isdir": lambda v=isdir: v,
                    "isfile": lambda v=isdir: not isdir,  # noqa: B023
                },
            )
            for name, isdir in archive_items
        ]
        tar_mock = flexmock(getmembers=lambda: tarinfo_mock_list)
        tarfile_mock.should_receive("open").and_return(tar_mock)
        tarfile_mock.should_receive("is_tarfile").and_return(is_tarfile)

    return tar_mock_factory


@pytest.mark.parametrize(
    "fork_username",
    [
        pytest.param("test_fork_username", id="fork_username_set"),
        pytest.param(None, id="fork_username_None"),
    ],
)
def test_create_pull(upstream_mock, upstream_pr_mock, fork_username):
    upstream_mock.local_project.git_project.should_receive("create_pr").with_args(
        title="test_title",
        body="test_description",
        source_branch="test_source",
        target_branch="test_target",
        fork_username=fork_username,
    ).and_return(upstream_pr_mock)
    upstream_mock.create_pull(
        pr_title="test_title",
        pr_description="test_description",
        source_branch="test_source",
        target_branch="test_target",
        fork_username=fork_username,
    )


@pytest.mark.parametrize(
    "action_config,result",
    [
        pytest.param("one cmd", [["one", "cmd"]], id="str_command"),
        pytest.param(["one cmd"], [["one", "cmd"]], id="list_command"),
        pytest.param([["one", "cmd"]], [["one", "cmd"]], id="list_in_list_command"),
        pytest.param(
            ["one cmd", "second cmd"],
            [["one", "cmd"], ["second", "cmd"]],
            id="two_str_commands_in_list",
        ),
        pytest.param(
            [["one", "cmd"], ["second", "cmd"]],
            [["one", "cmd"], ["second", "cmd"]],
            id="two_list_commands_in_list",
        ),
        pytest.param(
            [["one", "cmd"], "second cmd"],
            [["one", "cmd"], ["second", "cmd"]],
            id="one_str_and_one_list_command_in_list",
        ),
    ],
)
def test_get_commands_for_actions(action_config, result):
    ups = GitUpstream(
        package_config=flexmock(
            actions={ActionName.create_archive: action_config},
            synced_files=flexmock(),
        ),
        config=flexmock(),
        local_project=flexmock(),
    )
    ups._command_handler = flexmock()
    assert (
        ups.actions_handler.get_commands_for_actions(action=ActionName.create_archive)
        == result
    )


@pytest.mark.parametrize(
    "action_output, version, expected_result",
    [
        pytest.param(
            ("some_action_output", "1.0.1"),
            "_",
            "1.0.1",
            id="with_action_output",
        ),
        pytest.param(None, "1.0.2", "1.0.2", id="tag_valid_version"),
        pytest.param(None, "1.0-3", "1.0.3", id="tag_version_with_dash"),
    ],
)
def test_get_current_version(action_output, version, expected_result, upstream_mock):
    flexmock(packit.upstream.os).should_receive("listdir").and_return("mocked")
    flexmock(ActionsHandler).should_receive("get_output_from_action").and_return(
        action_output,
    )
    upstream_mock.should_receive("get_last_tag").and_return("_mocked")
    upstream_mock.should_receive("get_version_from_tag").and_return(version)
    upstream_mock.package_config.should_receive("get_package_names_as_env").and_return(
        {},
    )
    assert upstream_mock.get_current_version() == expected_result


@pytest.mark.parametrize(
    "tag, tag_template, expected_output, expectation",
    [
        pytest.param(
            "1.0.0",
            "{version}",
            "1.0.0",
            does_not_raise(),
            id="pure_version-valid_template",
        ),
        pytest.param(
            "test-1.0.0",
            "test-{version}",
            "1.0.0",
            does_not_raise(),
            id="valid_string-valid_template",
        ),
        pytest.param(
            "1.0.0",
            "tag-{random_field}",
            "1.0.0",
            pytest.raises(PackitException),
            id="missing_version_in_template",
        ),
        pytest.param(
            "some_name-1.0.0",
            "other_name-{version}",
            "1.0.0",
            pytest.raises(PackitException),
            id="no_match_found",
        ),
    ],
)
def test_get_version_from_tag(
    tag,
    tag_template,
    expected_output,
    expectation,
    upstream_mock,
):
    with expectation:
        upstream_mock.package_config.upstream_tag_template = tag_template
        assert upstream_mock.get_version_from_tag(tag) == expected_output


@pytest.mark.parametrize(
    "archive_type, return_value",
    [
        pytest.param("tar", "inner_archive_dir", id="tar_archive"),
        pytest.param("unknown", "dir_from_template", id="unknown_archive"),
    ],
)
def test_get_archive_root_dir(archive_type, return_value, upstream_mock, tar_mock):
    if archive_type == "tar":
        tar_mock(is_tarfile=True)
        flexmock(Archive).should_receive("get_archive_root_dir_from_tar").and_return(
            return_value,
        ).with_args("_archive").once()
        assert (
            Archive(upstream_mock, "").get_archive_root_dir("_archive") == return_value
        )
    elif archive_type == "unknown":
        flexmock(Archive).should_receive(
            "get_archive_root_dir_from_template",
        ).and_return(return_value)
        tar_mock(is_tarfile=False)
        assert (
            Archive(upstream_mock, "").get_archive_root_dir("_archive") == return_value
        )


@pytest.mark.parametrize(
    "archive_items, expected_result",
    [
        pytest.param(
            [("dir1", True), ("dir1/dir2", True), ("dir1/file1", False)],
            "dir1",
            id="valid_archive",
        ),
        pytest.param(
            [("dir1/dir2", True), ("dir1/file1", False)],
            "dir1",
            id="valid_archive_no_separate_top_level",
        ),
        pytest.param([], None, id="invalid_archive_empty"),
        pytest.param(
            [("dir1", True), ("dir2", True), ("dir1/file1", False)],
            None,
            id="invalid_two_dirs",
        ),
        pytest.param(
            [("dir1", True), ("dir1/dir2", True), ("file1", False)],
            "dir1",
            id="warning_file_in_root",
        ),
    ],
)
def test_get_tar_archive_dir(archive_items, expected_result, upstream_mock, tar_mock):
    tar_mock(archive_items=archive_items)
    assert (
        Archive(upstream_mock).get_archive_root_dir_from_tar("_archive")
        == expected_result
    )


@pytest.mark.parametrize(
    "template, expected_return_value",
    [
        pytest.param(
            "{upstream_pkg_name}-{version}",
            "test_package_name-1.0",
            id="default",
        ),
        pytest.param(
            "{version}-{upstream_pkg_name}",
            "1.0-test_package_name",
            id="custom",
        ),
        pytest.param("{unknown}-{version}", "{unknown}-1.0", id="unknown_tag"),
        pytest.param("static_string", "static_string", id="static_template"),
    ],
)
def test_get_archive_root_dir_from_template(
    template,
    expected_return_value,
    upstream_mock,
):
    upstream_mock.package_config.archive_root_dir_template = template
    assert (
        Archive(upstream_mock, "1.0").get_archive_root_dir_from_template()
        == expected_return_value
    )


@pytest.mark.parametrize(
    "version, tag_template, expected_output, expectation",
    [
        pytest.param(
            "1.0.0",
            "{version}",
            "1.0.0",
            does_not_raise(),
            id="valid_template",
        ),
        pytest.param(
            "1.0.0",
            "{rsion}",
            "1.0.0",
            pytest.raises(PackitException),
            id="invalid_template",
        ),
    ],
)
def test_convert_version_to_tag(
    version,
    tag_template,
    expected_output,
    expectation,
    upstream_mock,
):
    with expectation:
        upstream_mock.package_config.upstream_tag_template = tag_template
        assert upstream_mock.convert_version_to_tag(version) == expected_output


@pytest.mark.parametrize(
    "output, expected",
    (
        (
            "Wrote: packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm",
            [
                "packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm",
            ],
        ),
        (
            "Processing files: python3-packit-0.37.1.dev14+g860168a.d20211004-1."
            "20211004105435567001.rpm.regex.broken.14.g860168a.fc35.noarch\n"
            "\nAnother false positive: random_rpm_named_with_space .rpm"
            "Wrote: packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm",
            [
                "packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm",
            ],
        ),
        (
            "Wrote: packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm"
            "\n\n\nWrote: packit-0.38.0.rpma.fc35.noarch.rpm\n"
            "Wrote: packit-0.38.0.rpm_hmm.fc35.noarch.rpm\n",
            [
                "packit-0.37.1.dev13+gd57da48.rpm.regex.broken.13.gd57da48.fc35.noarch.rpm",
                "packit-0.38.0.rpma.fc35.noarch.rpm",
                "packit-0.38.0.rpm_hmm.fc35.noarch.rpm",
            ],
        ),
    ),
)
def test_get_rpms_from_rpmbuild_output(output, expected):
    assert GitUpstream._get_rpms_from_rpmbuild_output(output) == expected


@pytest.mark.parametrize(
    ("archive", "version", "release_suffix", "expanded_release_suffix"),
    [
        ("archive.tar.gz", "1.0.0", "123", "123"),
        # not allowed outside of tests, but the SHA is hardcoded for local project
        ("ravl.tar.bz2", "1.2.3", "{PACKIT_PROJECT_COMMIT}", "_"),
        ("one_piece.tar.gz", "1.0.2", "{PACKIT_PROJECT_BRANCH}", "mock_ref"),
        (
            "something.zip",
            "1.0.2",
            "{PACKIT_PROJECT_BRANCH}.{PACKIT_PROJECT_VERSION}",
            "mock_ref.1.0.2",
        ),
    ],
)
def test_release_suffix(
    upstream_mock,
    archive,
    version,
    release_suffix,
    expanded_release_suffix,
):
    flexmock(upstream_mock).should_receive("get_current_version").and_return(version)
    flexmock(upstream_mock).should_receive("get_spec_release").and_return(
        expanded_release_suffix,
    )
    upstream_mock.package_config.should_receive("get_package_names_as_env").and_return(
        {},
    )
    flexmock(upstream_mock).should_receive("specfile").and_return(
        flexmock(expanded_release=expanded_release_suffix)
        .should_receive("reload")
        .and_return(None)
        .mock(),
    )

    flexmock(upstream_mock).should_receive("fix_spec").with_args(
        archive=archive,
        version=version,
        commit="_",
        update_release=True,
        release=expanded_release_suffix,
    )

    SRPMBuilder(upstream_mock)._fix_specfile_to_use_local_archive(
        archive=archive,
        update_release=True,
        release_suffix=release_suffix,
    )


@pytest.mark.parametrize(
    "rpmbuild_output",
    (
        pytest.param(
            "setting SOURCE_DATE_EPOCH=1668038400"
            "Wrote: ./koji-c-1.1.0-1.20221110110837054647.pr257.1.g8477d03.fc35.src.rpm"
            "RPM build warnings:"
            "    extra tokens at the end of %endif directive in line 66:  %endif # with_python3",
            id="output_after",
        ),
        pytest.param(
            "setting SOURCE_DATE_EPOCH=1668038400"
            "Wrote: ./koji-c-1.1.0-1.20221110110837054647.pr257.1.g8477d03.fc35.src.rpm",
            id="common_output",
        ),
    ),
)
def test_get_srpm_from_rpmbuild_output(upstream_mock, rpmbuild_output):
    srpm = "./koji-c-1.1.0-1.20221110110837054647.pr257.1.g8477d03.fc35.src.rpm"
    assert srpm == SRPMBuilder(upstream_mock)._get_srpm_from_rpmbuild_output(
        rpmbuild_output,
    )


@pytest.mark.parametrize(
    "update_release,release_suffix,expected_release",
    (
        # current_git_tag_version="4.5"
        # original_release_number_from_spec = "2"
        pytest.param(
            True,
            "",
            # update-release from command line wins over release_suffix on packit.yaml
            "2.1234.mock_ref.",
            id="Bump release, release_suffix is empty",
        ),
        pytest.param(
            True,
            None,
            "2.1234.mock_ref.",
            id="Bump release, release_suffix is None",
        ),
        pytest.param(True, "7", "2.7", id="Bump release, release_suffix value is 7"),
        pytest.param(
            True,
            "{PACKIT_RPMSPEC_RELEASE}",
            "2.1234.mock_ref.",
            id="Bump release, release_suffix value is a macro {PACKIT_RPMSPEC_RELEASE}",
        ),
        pytest.param(
            False,
            "",
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix is empty",
        ),
        pytest.param(
            False,
            None,
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix is None",
        ),
        pytest.param(
            False,
            "7",
            "2.7",
            id="Do not modify release, release_suffix value is 7",
        ),
        pytest.param(
            False,
            "{PACKIT_RPMSPEC_RELEASE}",
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix value is a macro {PACKIT_RPMSPEC_RELEASE}",
        ),
    ),
)
def test_get_spec_release(
    upstream_mock,
    update_release,
    release_suffix,
    expected_release,
):
    archive = "an_archive_name"
    current_git_tag_version = "4.5"
    original_release_number_from_spec = "2"
    flexmock(upstream_mock).should_receive("get_current_version").and_return(
        current_git_tag_version,
    )
    upstream_mock.package_config.should_receive("get_package_names_as_env").and_return(
        {},
    )

    flexmock(sys.modules["packit.upstream"]).should_receive("datetime").and_return(
        flexmock(datetime=flexmock(now=flexmock(strftime=lambda f: "1234"))),
    )

    flexmock(upstream_mock).should_receive("fix_spec").with_args(
        archive=archive,
        version=current_git_tag_version,
        commit="_",
        update_release=update_release,
        release=expected_release,
    )
    upstream_mock._specfile = flexmock(
        expanded_release=original_release_number_from_spec,
    )
    upstream_mock._specfile.should_receive("reload").once()

    flexmock(sys.modules["packit.upstream"]).should_receive("run_command").and_return(
        flexmock(stdout=current_git_tag_version),
    )

    SRPMBuilder(upstream_mock)._fix_specfile_to_use_local_archive(
        archive=archive,
        update_release=update_release,
        release_suffix=release_suffix,
    )


@pytest.mark.parametrize(
    "update_release,had_disttag,release_suffix,expected_release_suffix",
    (
        # current_git_tag_version="4.5"
        # original_release_number_from_spec = "2"
        # original_dist_from_spec = "%{?dist}" if had_disttag else ""
        pytest.param(
            True,
            False,
            "",
            # update-release from command line wins over release_suffix on packit.yaml
            "2.1234.mock_ref.%{?dist}",
            id="Bump release, release_suffix is empty",
        ),
        pytest.param(
            True,
            False,
            None,
            "2.1234.mock_ref.%{?dist}",
            id="Bump release, release_suffix is None",
        ),
        pytest.param(
            True,
            False,
            "7",
            "2.7%{?dist}",
            id="Bump release, release_suffix value is 7",
        ),
        pytest.param(
            True,
            False,
            "{PACKIT_RPMSPEC_RELEASE}",
            "2.1234.mock_ref.%{?dist}",
            id="Bump release, release_suffix value is a macro {PACKIT_RPMSPEC_RELEASE}",
        ),
        pytest.param(
            True,
            True,
            "",
            "2.1234.mock_ref.",
            id="Bump release, release_suffix is empty, make sure %{?dist} tag is not duplicated",
        ),
        pytest.param(
            True,
            True,
            "{PACKIT_RPMSPEC_RELEASE}",
            "2.1234.mock_ref.",
            id="Bump release, release_suffix is a macro, make sure %{?dist} tag is not duplicated",
        ),
        pytest.param(
            True,
            False,
            "{PACKIT_RPMSPEC_RELEASE}%{{?dist}}",
            "2.1234.mock_ref.%{?dist}",
            id="Bump release, make sure %{?dist} tag is not duplicated",
        ),
        pytest.param(
            False,
            False,
            "",
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix is empty",
        ),
        pytest.param(
            False,
            False,
            None,
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix is None",
        ),
        pytest.param(
            False,
            False,
            "7",
            "2.7",
            id="Do not modify release, release_suffix value is 7",
        ),
        pytest.param(
            False,
            False,
            "{PACKIT_RPMSPEC_RELEASE}",
            "2.1234.mock_ref.",
            id="Do not modify release, release_suffix value is a macro {PACKIT_RPMSPEC_RELEASE}",
        ),
    ),
)
def test_fix_spec(
    upstream_mock,
    update_release,
    had_disttag,
    release_suffix,
    expected_release_suffix,
):
    upstream_mock.package_config.upstream_tag_include = None
    upstream_mock.package_config.upstream_tag_exclude = None
    upstream_mock.package_config.should_receive("get_package_names_as_env").and_return(
        {},
    )
    archive = "an_archive_name"
    current_git_tag_version = "4.5"
    original_release_number_from_spec = "2"
    flexmock(upstream_mock).should_receive("get_current_version").and_return(
        current_git_tag_version,
    )

    flexmock(sys.modules["packit.upstream"]).should_receive("datetime").and_return(
        flexmock(datetime=flexmock(now=flexmock(strftime=lambda f: "1234"))),
    )

    flexmock(upstream_mock).should_receive("_fix_spec_source").with_args(
        archive=archive,
    )
    flexmock(upstream_mock).should_receive("_fix_spec_prep").with_args(archive=archive)

    disttag = "%{?dist}" if had_disttag else ""
    upstream_mock._specfile = flexmock(
        expanded_release=original_release_number_from_spec,
        raw_release=f"{original_release_number_from_spec}{disttag}",
    )
    upstream_mock._specfile.should_receive("reload").once()

    if update_release:
        upstream_mock._specfile.should_receive("add_changelog_entry")

    flexmock(sys.modules["packit.upstream"]).should_receive("run_command").and_return(
        flexmock(stdout=current_git_tag_version),
    )

    SRPMBuilder(upstream_mock)._fix_specfile_to_use_local_archive(
        archive=archive,
        update_release=update_release,
        release_suffix=release_suffix,
    )

    assert upstream_mock._specfile.version == current_git_tag_version

    if update_release:
        assert upstream_mock._specfile.release == expected_release_suffix



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from subprocess import CalledProcessError, check_output

import packit.distgit
import packit.upstream
from packit.config import Config, get_package_config_from_repo
from packit.exceptions import PackitException
from packit.local_project import CALCULATE, LocalProjectBuilder


def socket_guard(*args, **kwargs):
    raise Exception("Internet connection not allowed")


class PackitTest(unittest.TestCase):
    def setUp(self):
        self._static_tmp = None
        self._config = None
        self._project = None
        # if you can, use this project to perform testing
        self._project_url = "https://github.com/packit/requre"
        self._project_specfile_path = Path("fedora", "python-requre.spec")
        self._pc = None
        self._dg = None
        self._lp = None
        self._upstream = None

    def tearDown(self):
        if self._static_tmp and os.path.exists(self._static_tmp):
            shutil.rmtree(self.static_tmp, ignore_errors=True)
            self._static_tmp = None

    @property
    def static_tmp(self):
        if not self._static_tmp:
            self._static_tmp = tempfile.mkdtemp()
            os.makedirs(self._static_tmp, exist_ok=True)
        return self._static_tmp

    @property
    def config(self):
        if not self._config:
            try:
                self._config = Config.get_user_config()
            except PackitException:
                self._config = Config()
        return self._config

    @property
    def project(self):
        if not self._project:
            self._project = self.config.get_project(url=self._project_url)
        return self._project

    @property
    def pc(self):
        if not self._pc:
            self._pc = get_package_config_from_repo(project=self.project)
        if not self._pc:
            raise RuntimeError("Package config not found.")
        return self._pc

    @property
    def dg(self):
        if not self._dg:
            self._dg = packit.distgit.DistGit(self.config, self.pc)
        return self._dg

    @property
    def lp(self):
        if not self._lp:
            # Fully construct LocalProject, the tests may need anything
            builder = LocalProjectBuilder()
            self._lp = builder.build(
                working_dir=CALCULATE,
                ref=CALCULATE,
                git_repo=CALCULATE,
                git_project=self.project,
                git_service=CALCULATE,
                full_name=CALCULATE,
                namespace=CALCULATE,
                repo_name=CALCULATE,
                git_url=CALCULATE,
            )
        return self._lp

    @property
    def upstream(self):
        if not self._upstream:
            self._upstream = packit.upstream.GitUpstream(
                self.config,
                self.pc,
                local_project=self.lp,
            )
        return self._upstream

    @property
    def project_specfile_location(self):
        return self.lp.working_dir / self._project_specfile_path

    @staticmethod
    def configure_git():
        try:
            check_output(["git", "config", "--global", "-l"])
        except CalledProcessError:
            check_output(
                ["git", "config", "--global", "user.email", "test@example.com"],
            )
            check_output(["git", "config", "--global", "user.name", "Tester"])
            check_output(["git", "config", "--global", "safe.directory", "*"])



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import dataclasses
import logging
import shutil
from collections.abc import Iterable
from functools import partial
from pathlib import Path
from typing import Any, ClassVar, Optional, TypeVar, Union

import git
from git.exc import GitCommandError
from ogr import GitlabService
from ogr.abstract import GitProject, GitService
from ogr.factory import get_project, get_service_class
from ogr.parsing import parse_git_repo

from packit.constants import LP_TEMP_PR_CHECKOUT_NAME
from packit.exceptions import PackitException, PackitMergeException
from packit.utils.repo import (
    RepositoryCache,
    get_repo,
    is_git_repo,
    shorten_commit_hash,
)

logger = logging.getLogger(__name__)


class LocalProject:
    """
    Class representing a cloned repository
    and its API to the remote git-forge (e.g. GitHub/GitLab/Pagure)

    - git_repo: instance of git.Repo
    - working_dir: working directory for the project
    - ref: git ref (branch/tag/commit) if set, then checkouted
    - git_project: instance of ogr.GitProject (remote API for project)
    - git_service: instance of ogr.GitService (tokens for remote API)
    - git_url: remote url (used for cloning)
    - full_name: "$namespace/$repo"
    - namespace: namespace of the remote project
    - repo_name: name of the remote project


    Local project can compute other attributes if it is possible.
    """

    # setting defaults to str because `None == ""` results into TypeError is not true-true
    def __init__(
        self,
        git_repo: Optional[git.Repo] = None,
        working_dir: Union[Path, str, None] = None,
        ref: str = "",
        git_project: Optional[GitProject] = None,
        git_service: Optional[GitService] = None,
        git_url: str = "",
        full_name: str = "",
        namespace: str = "",
        repo_name: str = "",
        offline: bool = False,
        refresh: bool = True,
        remote: str = "",
        pr_id: Optional[str] = None,
        cache: Optional[RepositoryCache] = None,
        merge_pr: bool = True,
        target_branch: str = "",
        working_dir_temporary: bool = False,
    ) -> None:
        """

        :param git_repo: git.Repo
        :param working_dir: Path|str (working directory for the project)
        :param ref: str (git ref (branch/tag/commit) if set, then checked out)
        :param git_project: ogr.GitProject (remote API for project)
        :param git_service: ogr.GitService (tokens for remote API)
        :param git_url: str (remote url used for cloning)
        :param full_name: str ("$namespace/$repo")
        :param namespace: str (namespace of the remote project)
        :param repo_name: str (name of the remote project)
        :param offline: bool (do not use any network action, defaults to False)
        :param refresh: bool (calculate the missing attributes, defaults to True)
        :param remote: name of the git remote to use
        :param pr_id: ID of the pull request to fetch and check out
        """
        self.working_dir_temporary = working_dir_temporary
        self._git_repo: git.Repo = git_repo
        self.working_dir: Optional[Path] = Path(working_dir) if working_dir else None
        self._ref = ref
        self.git_project = git_project
        self.git_service = git_service
        self.git_url = git_url
        self.full_name = full_name
        self.repo_name = repo_name
        self.namespace = namespace
        self.offline = offline
        self.remote = remote
        self.cache = cache

        logger.debug(
            "Arguments received in the init method of the LocalProject class: \n"
            f"git_repo: {git_repo}\n"
            f"working_dir: {working_dir}\n"
            f"ref: {ref}\n"
            f"git_project: {git_project}\n"
            f"git_service: {git_service}\n"
            f"git_url: {git_url}\n"
            f"full_name: {full_name}\n"
            f"namespace: {namespace}\n"
            f"repo_name: {repo_name}\n"
            f"offline: {offline}\n"
            f"refresh {refresh}\n"
            f"remote: {remote}\n"
            f"pr_id: {pr_id}\n"
            f"cache: {cache}\n"
            f"merge_pr: {merge_pr}\n"
            f"target_branch: {target_branch}\n",
        )

        if refresh:
            self.refresh_the_arguments()

        # skip checkouts if the git repo is not present
        if not self.git_repo:
            return

        # p-s gives us both, commit hash for a PR and PR ID as well
        # since we want to have 'pr123' in the release field, let's check out
        # the PR itself, so if both are specified, PR ID > ref
        if pr_id:
            self.checkout_pr(pr_id)
            if merge_pr:
                self.merge_pr(pr_id, target_branch)
            self.checkout_as_pr_branch(pr_id)
        elif ref:
            self.checkout_ref(ref)

    def __repr__(self):
        return (
            "LocalProject("
            f"working_dir_temporary='{self.working_dir_temporary}', "
            f"git_repo='{self.git_repo}', "
            f"working_dir='{self.working_dir}', "
            f"ref='{self.ref}', "
            f"git_project='{self.git_project}', "
            f"git_service='{self.git_service}', "
            f"git_url='{self.git_url}', "
            f"full_name='{self.full_name}', "
            f"repo_name='{self.repo_name}', "
            f"namespace='{self.namespace}', "
            f"offline='{self.offline}', "
            f"remote='{self.remote}', "
            f"commit_hexsha='{self.commit_hexsha}')"
        )

    @property
    def git_repo(self) -> Optional[git.Repo]:
        """The git.Repo tied to this LocalProject.

        This is useful for creating class-level mocks with flexmock rather
        than having to worry about instantiation of the LocalProject class.
        The property can be mocked so that it returns a mock repo with the
        required capabilities.
        """
        return self._git_repo

    @property
    def ref(self) -> Optional[str]:
        """
        Name of the HEAD if the HEAD is not detached,
        else commit hash.
        """
        return self._get_ref_from_git_repo() if self.git_repo else None

    @property
    def commit_hexsha(self) -> Optional[str]:
        """
        Get the short commit hash for the current commit.

        :return: first 8 characters of the current commit
        """
        if not self.git_repo:
            return None
        if self.git_repo.head.is_detached:
            return shorten_commit_hash(self.git_repo.head.commit.hexsha)
        return shorten_commit_hash(self.git_repo.active_branch.commit.hexsha)

    def clean(self):
        """Remove the git tree when cloned into a temporary directory"""
        if self.working_dir_temporary:
            logger.debug(f"Cleaning: {self.working_dir}")
            shutil.rmtree(self.working_dir, ignore_errors=True)
            self.working_dir_temporary = False

    def free_resources(self):
        """Clean internal git cache which GitPython uses in the background, suggested solution:
        https://github.com/gitpython-developers/GitPython/issues/546#issuecomment-256657166
        the code of the function clearly manipulates the git-cat-file operations
        which we are seeing hang (source: git.cmd.Git.clear_cache)
        """
        if self.git_repo:  # tests in p-s can have `self.git_repo = None`
            self.git_repo.git.clear_cache()

    def refresh_the_arguments(self):
        # TODO: remove this whole logic once everything uses LocalProjectBuilder
        change = True
        while change:
            # we are trying to get new information while it is possible
            # new iteration is done only if there was a change in the last iteration

            change = (
                self._parse_repo_name_full_name_and_namespace()
                or self._parse_git_repo_from_working_dir()
                or self._parse_git_project_from_repo_namespace_and_git_service()
                or self._parse_git_service_from_git_project()
                or self._parse_ref_from_git_repo()
                or self._parse_working_dir_from_git_repo()
                or self._parse_git_repo_from_git_url()
                or self._parse_git_url_from_git_project()
                or self._parse_repo_name_from_git_project()
                or self._parse_namespace_from_git_project()
                or self._parse_git_url_from_git_repo()
                or self._parse_namespace_from_git_url()
            )

    def _parse_repo_name_full_name_and_namespace(self):
        change = False
        if self.repo_name and self.namespace and not self.full_name:
            self.full_name = f"{self.namespace}/{self.repo_name}"
            change = True
        if self.full_name and not self.namespace:
            self.namespace = self.full_name.split("/")[0]
            change = True
        if self.full_name and not self.repo_name:
            self.repo_name = self.full_name.split("/")[1]
            change = True

        if change:
            logger.debug(f"Parsed full repo name '{self.namespace}/{self.repo_name}'.")
        return change

    def _parse_git_repo_from_working_dir(self) -> bool:
        """
        Get the repo from the self.working_dir (clone self.git_url if it is not a git repo)
        """
        if self.working_dir and not self.git_repo:
            logger.debug(
                "`working_dir` is set and `git_repo` is not: let's discover...",
            )
            if is_git_repo(directory=self.working_dir):
                logger.debug("It's a git repo!")
                self._git_repo = git.Repo(path=self.working_dir)
                return True

            if self.git_url and not self.offline:
                self._git_repo = self._get_repo(
                    url=self.git_url,
                    directory=self.working_dir,
                )
                logger.debug(
                    f"We just cloned git repo {self.git_url} to {self.working_dir}.",
                )
                return True
        return False

    def _parse_git_project_from_repo_namespace_and_git_service(
        self,
    ) -> bool:
        if (
            self.repo_name
            and self.namespace
            and self.git_service
            and not self.git_project
            and not self.offline
        ):
            self.git_project = self.git_service.get_project(
                repo=self.repo_name,
                namespace=self.namespace,
            )
            logger.debug(f"Parsed project '{self.namespace}/{self.repo_name}'.")
            return True
        return False

    def _parse_git_service_from_git_project(self):
        if not (self.git_project is None or self.git_service or self.offline):
            self.git_service = self.git_project.service
            logger.debug(
                f"Parsed service {self.git_service} from the project {self.git_project}.",
            )
            return True
        return False

    def _parse_ref_from_git_repo(self):
        if self.git_repo and not self._ref:
            self._ref = self._get_ref_from_git_repo()
            logger.debug(f"Parsed ref {self._ref!r} from the repo {self.git_repo}.")
            return bool(self._ref)
        return False

    def _parse_working_dir_from_git_repo(self):
        if self.git_repo and not self.working_dir:
            self.working_dir = Path(self.git_repo.working_dir)
            logger.debug(
                f"Parsed working directory {self.working_dir} from the repo {self.git_repo}.",
            )
            return True
        return False

    def _parse_git_repo_from_git_url(self):
        if (
            self.git_url
            and not self.working_dir
            and not self.git_repo
            and not self.offline
        ):
            self._git_repo = self._get_repo(url=self.git_url)
            self.working_dir_temporary = True
            logger.debug(f"Parsed repo {self.git_repo} from url {self.git_url!r}.")
            return True
        return False

    def _parse_git_url_from_git_project(self):
        if self.git_project and not self.git_url and not self.offline:
            self.git_url = self.git_project.get_git_urls()["git"]
            logger.debug(
                f"Parsed remote url {self.git_url!r} from the project {self.git_project}.",
            )
            return True
        return False

    def _parse_repo_name_from_git_project(self):
        if self.git_project and not self.repo_name:
            self.repo_name = self.git_project.repo
            if not self.repo_name:
                raise PackitException(
                    "Repo name should have been set but isn't, this is bug!",
                )
            logger.debug(
                f"Parsed repo name {self.repo_name!r} from the git project {self.git_project}.",
            )
            return True
        return False

    def _parse_namespace_from_git_project(self):
        if self.git_project and not self.namespace:
            self.namespace = self.git_project.namespace
            logger.debug(
                f"Parsed namespace {self.namespace!r} from the project {self.git_project}.",
            )
            return True
        return False

    def _parse_git_url_from_git_repo(self):
        if not self.git_repo or self.git_url:
            return False

        if self.remote:
            self.git_url = next(self.git_repo.remote(self.remote).urls)
        elif self.git_repo.remotes:
            for remote in self.git_repo.remotes:
                if remote.name == "origin":
                    # origin as a default
                    self.git_url = remote.url
                    break
            else:
                # or use first one
                self.git_url = next(self.git_repo.remotes[0].urls)
        else:
            # Repo has no remotes
            return False
        logger.debug(
            f"Parsed remote url {self.git_url!r} from the repo {self.git_repo}.",
        )
        return True

    def _parse_namespace_from_git_url(self):
        if self.git_url and not (self.namespace and self.repo_name):
            parsed_repo_url = parse_git_repo(potential_url=self.git_url)
            if (
                parsed_repo_url.namespace == self.namespace
                and parsed_repo_url.repo == self.repo_name
            ):
                return False
            self.namespace, self.repo_name = (
                parsed_repo_url.namespace,
                parsed_repo_url.repo,
            )
            logger.debug(
                f"Parsed namespace and repo name ({self.namespace}, {self.repo_name}) "
                f"from url {self.git_url!r}.",
            )
            return True
        return False

    def _get_ref_from_git_repo(self) -> str:
        if self.git_repo.head.is_detached:
            return self.git_repo.head.commit.hexsha
        return self.git_repo.active_branch.name

    def _get_repo(self, url, directory=None):
        if self.cache:
            return self.cache.get_repo(url, directory=directory)
        return get_repo(url=url, directory=directory)

    def checkout_ref(self, ref: str):
        """Check out selected ref in the git repo"""
        logger.info(f"Checking out ref {ref!r}.")
        self.git_repo.git.checkout(ref)
        logger.debug(f"Current commit is '{self.git_repo.commit()}'")

    def create_branch(
        self,
        branch_name: str,
        base: str = "HEAD",
        setup_tracking: bool = False,
    ) -> git.Head:
        """
        Create a new git branch in git

        :param branch_name: name of the branch to check out and fetch
        :param base: we base our new branch on this one
        :param setup_tracking: set up remote tracking
               (exc will be raised if the branch is not in the remote)
        :return the branch which was just created
        """
        # it's not an error if the branch already exists
        if branch_name in self.git_repo.branches:
            logger.debug(
                f"It seems that branch {branch_name!r} already exists, checking it out.",
            )
            head = self.git_repo.branches[branch_name]
        else:
            head = self.git_repo.create_head(branch_name, commit=base)

        if setup_tracking:
            origin = self.git_repo.remote("origin")
            if branch_name in origin.refs:
                remote_ref = origin.refs[branch_name]
            else:
                raise PackitException(
                    f"Remote origin doesn't have ref {branch_name!r}.",
                )
            # this is important to fedpkg: build can't find the tracking branch otherwise
            head.set_tracking_branch(remote_ref)

        return head

    def checkout_as_pr_branch(self, pr_id: Union[str, int]) -> None:
        """
        Rename current branch into pr/{pr_id}.

        Args:
            pr_id: ID of the PR we are merging.
        """
        branch = self.git_repo.create_head(f"pr/{pr_id}", "HEAD")
        branch.checkout()

    def _fetch_as_branch(
        self,
        remote_ref: str,
        local_ref: str,
        local_branch: str,
    ) -> None:
        """
        Fetches reference from the remote as the specified local reference and
        creates a branch for it.

        Args:
            remote_ref: Git reference to be fetched from remote.
            local_ref: Git reference that refers to the remote reference.
            local_branch: Branch that represents local reference.
        """
        remote = self.remote or "origin"
        self.git_repo.remotes[remote].fetch(f"{remote_ref}:{local_ref}")
        # overwrite the local checkout when needed, remote is always accurate
        self.git_repo.create_head(local_branch, f"{remote}/{local_branch}", force=True)

    def checkout_pr(self, pr_id: Union[str, int]) -> None:
        """
        Fetch selected PR and check it out in a local branch `pr/{pr_id}`.

        Args:
            pr_id: ID of the PR we are merging.
        """
        logger.info(f"Checking out PR {pr_id}.")
        is_gitlab = isinstance(self.git_service, GitlabService) or (
            not self.git_service and get_service_class(self.git_url) == GitlabService
        )
        remote_ref = "+refs/{}/{}/head".format(
            "merge-requests" if is_gitlab else "pull",
            pr_id,
        )
        remote_name = self.remote or "origin"
        local_ref = f"refs/remotes/{remote_name}/{LP_TEMP_PR_CHECKOUT_NAME}/{pr_id}"
        local_branch = f"{LP_TEMP_PR_CHECKOUT_NAME}/{pr_id}"

        self._fetch_as_branch(remote_ref, local_ref, local_branch)
        self.git_repo.branches[local_branch].checkout()

        head_commit = self.git_repo.branches[local_branch].commit
        logger.info(
            f"Checked out commit\n"
            f"({shorten_commit_hash(head_commit.hexsha)})\t{head_commit.summary}",
        )

    def merge_pr(
        self,
        pr_id: Union[str, int],
        target_branch_name: Optional[str] = None,
    ) -> None:
        """
        Merge given PR into target branch. Fetches and switches to base branch
        (where changes from the PR are to be merged) and then merges branch with
        changes from the PR.

        Args:
            pr_id: ID of the PR we are merging.
            target_branch_name: name of the branch the PR should be merged into if
            git_project is None

        Raises:
            PackitException: In case merge fails.
        """
        remote = self.remote or "origin"
        target_branch_name = (
            self.git_project.get_pr(int(pr_id)).target_branch
            if self.git_project
            else target_branch_name
        )
        if not target_branch_name:
            raise PackitException(
                f"Cannot get the target branch for merging PR {pr_id}.",
            )

        logger.debug(f"Target branch: {target_branch_name}")

        self._fetch_as_branch(
            f"+refs/heads/{target_branch_name}",
            f"refs/remotes/{remote}/pr/{pr_id}",
            target_branch_name,
        )
        self.git_repo.branches[target_branch_name].checkout()
        target_branch = self.git_repo.branches[target_branch_name]

        commit_sha = shorten_commit_hash(target_branch.commit.hexsha)
        logger.info(
            f"Merging ({target_branch}) with commit:\n"
            f"({commit_sha})\t{target_branch.commit.summary}",
        )
        try:
            self.git_repo.git.merge(f"{LP_TEMP_PR_CHECKOUT_NAME}/{pr_id}")
        except GitCommandError as ex:
            logger.warning(f"Merge failed with: {ex}")
            if "Merge conflict" in str(ex):
                raise PackitMergeException(ex) from ex
            raise PackitException(ex) from ex

    def checkout_release(self, tag: str) -> None:
        logger.info(f"Checking out upstream version {tag}.")
        try:
            self.git_repo.git.checkout(tag)
        except Exception as ex:
            raise PackitException(
                "Cannot checkout release tag. Please, check whether the "
                "'upstream_tag_template' needs to be configured.",
            ) from ex

    def fetch(self, remote: str, refspec: Optional[str] = None, force: bool = False):
        """
        Fetch refs and/or tags from a remote to this repo.

        Args:
            remote: Str or path of the repo we fetch from.
            refspec: See man git-fetch.
            force: See --force in man git-fetch.
        """
        args = [remote]
        args += [refspec] if refspec else ["--tags"]
        if force:
            args += ["--force"]
        self.git_repo.git.fetch(*args)

    def __del__(self):
        self.clean()


@dataclasses.dataclass
class LocalProjectCalculationState:
    """Class representing the current state of construction of LocalProject.

    Encloses all the arguments that will then be passed to LocalProject's
    constructor.
    """

    remote: str = ""
    git_repo: Optional[git.Repo] = None
    working_dir: Union[Path, str, None] = None
    ref: str = ""
    git_project: Optional[GitProject] = None
    git_service: Optional[GitService] = None
    git_url: str = ""
    full_name: str = ""
    namespace: str = ""
    repo_name: str = ""
    working_dir_temporary: bool = False

    @classmethod
    def from_local_project(cls, lp: LocalProject) -> "LocalProjectCalculationState":
        """Constructs the calculation state from an existing LocalProject."""
        return cls(
            lp.remote,
            lp.git_repo,
            lp.working_dir,
            lp.ref,
            lp.git_project,
            lp.git_service,
            lp.git_url,
            lp.full_name,
            lp.namespace,
            lp.repo_name,
        )


# Use an explicit sentinel class to have stricter type-checking than just object()
class _CalculateType:
    pass


CALCULATE = _CalculateType()
NOT_TO_CALCULATE = _CalculateType()


T = TypeVar("T")
CanCalculate = Union[T, _CalculateType]


class LocalProjectBuilder:
    """Class for building instances of LocalProject dynamically."""

    PREREQUISITES: ClassVar = {
        "git_repo": ["working_dir", "git_url"],
        "git_project": ["git_url", "repo_name", "namespace", "git_service"],
        "git_service": ["git_project"],
        "ref": ["git_repo"],
        "working_dir": ["git_repo"],
        "git_url": ["git_project", "git_repo"],
        "repo_name": ["git_project", "full_name"],
        "namespace": ["git_project", "git_url", "full_name"],
        "full_name": ["namespace", "repo_name"],
    }

    def __init__(
        self,
        cache: Optional[RepositoryCache] = None,
        instances: Optional[Iterable[GitService]] = None,
        offline: bool = False,
    ):
        """Creates a builder instance.

        Args:
            cache: Repository cache that may be used for getting repos.
            instances: List of GitService instances to utilise while building.
            offline: Whether only offline operations should be performed in this builder.
        """
        self._instances = instances or []
        self._cache = cache
        self.offline = offline

    def _add_prerequisites_to_calculations(
        self,
        to_calculate: set[str],
        not_to_calculate: Optional[set[str]] = None,
    ) -> None:
        """Adds calculation prerequisites into to_calculate set.

        If a caller of this class requests git_repo to be calculated, they should not
        have to care about the prerequisites to git_repo, hence we add those to the
        calculation set as well.
        """
        logger.debug(f"Attributes requested: {', '.join(to_calculate)}")
        logger.debug(
            f"Attributes requested not to be calculated: {', '.join(not_to_calculate)}",
        )
        dependencies = {}
        change = True
        while change:
            len_before = len(to_calculate)
            for calc in to_calculate.copy():
                required = set(self.PREREQUISITES.get(calc, []))
                new_dependencies = required - to_calculate
                if not_to_calculate:
                    new_dependencies -= not_to_calculate
                to_calculate.update(new_dependencies)
                dependencies.update({calc: dep for dep in new_dependencies})
            change = len(to_calculate) > len_before

        dep_list = [f"{k} => {v}" for k, v in dependencies.items()]
        logger.debug(f"Transitive dependencies: {', '.join(dep_list)}")

    def _refresh_the_state(
        self,
        state: LocalProjectCalculationState,
        to_calculate: set[str],
        not_to_calculate: Optional[set[str]] = None,
    ) -> None:
        """Calculates the requested attributes while also considering transitive relations.

        Args:
            state: The initial state which will be updated with new calculated data.
            to_calculate: Set of attributes that need to be calculated.
        """
        self._add_prerequisites_to_calculations(to_calculate, not_to_calculate)
        # Remove the already set pieces
        to_calculate = set(filter(lambda a: not getattr(state, a, None), to_calculate))
        logger.debug(f"To-calculate set: {to_calculate}")
        name_partial = partial(self._parse_repo_name_full_name_and_namespace)
        all_partials = {
            "git_repo": [
                partial(self._parse_git_repo_from_git_url),
                partial(self._parse_git_repo_from_working_dir),
            ],
            "git_project": [
                partial(self._parse_git_project_from_url),
                partial(self._parse_git_project_from_repo_namespace_and_git_service),
            ],
            "git_service": [partial(self._parse_git_service_from_git_project)],
            "ref": [partial(self._parse_ref_from_git_repo)],
            "working_dir": [partial(self._parse_working_dir_from_git_repo)],
            "git_url": [
                partial(self._parse_git_url_from_git_project),
                partial(self._parse_git_url_from_git_repo),
            ],
            "repo_name": [
                name_partial,
                partial(self._parse_repo_name_from_git_project),
            ],
            "namespace": [
                name_partial,
                partial(self._parse_namespace_from_git_project),
                partial(self._parse_namespace_from_git_url),
            ],
            "full_name": [name_partial],
        }

        partials = [
            part for calc in to_calculate for part in all_partials.get(calc, [])
        ]
        change = True
        while change:
            change = False
            for part in partials:
                change = change or part(state)

    def _parse_repo_name_full_name_and_namespace(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Calculates repo name, namespace or full name if they are missing
        based on the other attributes."""
        change = False
        if state.repo_name and state.namespace and not state.full_name:
            state.full_name = f"{state.namespace}/{state.repo_name}"
            change = True
        if state.full_name and not state.namespace:
            state.namespace = state.full_name.split("/")[0]
            change = True
        if state.full_name and not state.repo_name:
            state.repo_name = state.full_name.split("/")[1]
            change = True

        if change:
            logger.debug(
                f"Parsed full repo name '{state.namespace}/{state.repo_name}'.",
            )
        return change

    def _parse_git_repo_from_working_dir(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Prepares git.Repo instance based on working_dir.

        Clones git_url if working_dir is not a repo.
        """
        if state.working_dir and not state.git_repo:
            logger.debug(
                "`working_dir` is set and `git_repo` is not: let's discover...",
            )
            if is_git_repo(directory=state.working_dir):
                logger.debug("It's a git repo!")
                state.git_repo = git.Repo(path=state.working_dir)
                return True

            if state.git_url and not self.offline:
                state.git_repo = self._get_repo(
                    url=state.git_url,
                    directory=state.working_dir,
                )
                logger.debug(
                    f"We just cloned git repo {state.git_url} to {state.working_dir}.",
                )
                return True

        return False

    def _parse_git_project_from_url(self, state: LocalProjectCalculationState) -> bool:
        """Creates GitProject based on git_url and provided GitService instances."""
        if state.git_url and self._instances:
            state.git_project = get_project(
                state.git_url,
                custom_instances=self._instances,
            )
            return True
        return False

    def _parse_git_project_from_repo_namespace_and_git_service(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Creates GitProject based on namespace, repo name and GitService."""

        if (
            state.repo_name
            and state.namespace
            and state.git_service
            and not state.git_project
            and not self.offline
        ):
            state.git_project = state.git_service.get_project(
                repo=state.repo_name,
                namespace=state.namespace,
            )
            logger.debug(f"Parsed project '{state.namespace}/{state.repo_name}'.")
            return True
        return False

    def _parse_git_service_from_git_project(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Creates GitService based on GitProject."""
        if not (state.git_project is None or state.git_service or self.offline):
            state.git_service = state.git_project.service
            logger.debug(
                f"Parsed service {state.git_service} from the project {state.git_project}.",
            )
            return True
        return False

    def _parse_ref_from_git_repo(self, state: LocalProjectCalculationState) -> bool:
        """Obtains the current git ref from git.Repo."""
        if state.git_repo and not state.ref:
            state.ref = self._get_ref_from_git_repo(state.git_repo)
            logger.debug(f"Parsed ref {state.ref!r} from the repo {state.git_repo}.")
            return bool(state.ref)
        return False

    def _parse_working_dir_from_git_repo(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Obtains the working_dir from git.Repo instance."""
        if state.git_repo and not state.working_dir:
            state.working_dir = Path(state.git_repo.working_dir)
            logger.debug(
                f"Parsed working directory {state.working_dir} from the repo {state.git_repo}.",
            )
            return True
        return False

    def _parse_git_repo_from_git_url(self, state: LocalProjectCalculationState) -> bool:
        """Prepares git.Repo instance based on a git URL."""
        if (
            state.git_url
            and not state.working_dir
            and not state.git_repo
            and not self.offline
        ):
            state.git_repo = self._get_repo(url=state.git_url)
            state.working_dir_temporary = True
            logger.debug(f"Parsed repo {state.git_repo} from url {state.git_url!r}.")
            return True
        return False

    def _parse_git_url_from_git_project(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Obtains git URL from a GitProject."""
        if state.git_project and not state.git_url and not self.offline:
            state.git_url = state.git_project.get_git_urls()["git"]
            logger.debug(
                f"Parsed remote url {state.git_url!r} from the project {state.git_project}.",
            )
            return True
        return False

    def _parse_repo_name_from_git_project(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Obtains git repo name from a GitProject."""
        if state.git_project and not state.repo_name:
            state.repo_name = state.git_project.repo
            if not state.repo_name:
                raise PackitException(
                    "Repo name should have been set but isn't, this is bug!",
                )
            logger.debug(
                f"Parsed repo name {state.repo_name!r} from the git project {state.git_project}.",
            )
            return True
        return False

    def _parse_namespace_from_git_project(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Obtains git repo namespace from a GitProject."""
        if state.git_project and not state.namespace:
            state.namespace = state.git_project.namespace
            logger.debug(
                f"Parsed namespace {state.namespace!r} from the project {state.git_project}.",
            )
            return True
        return False

    def _parse_git_url_from_git_repo(self, state: LocalProjectCalculationState) -> bool:
        """Obtains git URL from git.Repo based on the set remote."""
        if not state.git_repo or state.git_url:
            return False

        if state.remote:
            state.git_url = next(state.git_repo.remote(state.remote).urls)
        elif state.git_repo.remotes:
            for remote in state.git_repo.remotes:
                if remote.name == "origin":
                    # origin as a default
                    state.git_url = remote.url
                    break
            else:
                # or use first one
                state.git_url = next(state.git_repo.remotes[0].urls)
        else:
            # Repo has no remotes
            return False
        logger.debug(
            f"Parsed remote url {state.git_url!r} from the repo {state.git_repo}.",
        )
        return True

    def _parse_namespace_from_git_url(
        self,
        state: LocalProjectCalculationState,
    ) -> bool:
        """Obtains git repo namespace from a git URL."""
        if state.git_url and not (state.namespace and state.repo_name):
            parsed_repo_url = parse_git_repo(potential_url=state.git_url)
            if (
                parsed_repo_url.namespace == state.namespace
                and parsed_repo_url.repo == state.repo_name
            ):
                return False
            state.namespace, state.repo_name = (
                parsed_repo_url.namespace,
                parsed_repo_url.repo,
            )
            logger.debug(
                f"Parsed namespace and repo name ({state.namespace}, {state.repo_name}) "
                f"from url {state.git_url!r}.",
            )
            return True
        return False

    @staticmethod
    def _get_ref_from_git_repo(git_repo: git.Repo) -> str:
        if git_repo.head.is_detached:
            return git_repo.head.commit.hexsha
        return git_repo.active_branch.name

    def _get_repo(self, url, directory=None):
        if self._cache:
            return self._cache.get_repo(url, directory=directory)
        return get_repo(url=url, directory=directory)

    def build(
        self,
        # Non-building stuff
        remote: str = "",
        pr_id: Optional[str] = None,
        merge_pr: bool = True,
        target_branch: str = "",
        # Building data/instructions
        git_repo: CanCalculate[Union[git.Repo, None]] = None,
        working_dir: CanCalculate[Union[Path, str, None]] = None,
        ref: CanCalculate[str] = "",
        git_project: CanCalculate[Union[GitProject, None]] = None,
        git_service: CanCalculate[Union[GitService, None]] = None,
        git_url: CanCalculate[str] = "",
        full_name: CanCalculate[str] = "",
        namespace: CanCalculate[str] = "",
        repo_name: CanCalculate[str] = "",
        # LocalProject "template"
        local_project: Optional[LocalProject] = None,
    ) -> LocalProject:
        """Builds the LocalProject instance from the provided attributes.

        Most arguments can either take an explicit value, None (or empty string) or
        they can be passed the sentinel value CALCULATE to signal that this attribute
        must be present in the final LocalProject object and therefore be computed
        dynamically based on the rest of the information.

        Args:
            remote: Name of the git remote to use.
            pr_id: ID of the pull request to fetch and check out.
            merge_pr: Whether the fetched pull request should be merged.
            target_branch: Name of the branch the PR should be merged into.
            git_repo: Instance of git.Repo.
            working_dir: Path to the working directory of the project.
            ref: Git ref, if set, it will be checked out.
            git_project: ogr.GitProject (remote API)
            git_service: ogr.Gitservice (tokens for remote API)
            git_url: URL of the git project (used for cloning)
            full_name: Full name of the remote git project (including namespace).
            repo_name: Name of the remote git project.
            namespace: Namespace of the remote git project.
            local_project: LocalProject to use as the initial calculation state,
                the other arguments will override information present in this LocalProject.

        Returns:
            The constructed LocalProject instance.

        Examples:
            Clone the given URL and checkout the develop branch.

                builder.build(
                    git_url="https://github.com/packit/hello-world",
                    git_repo="CALCULATE",
                    ref="develop",
                )
        """
        to_calculate: set[str] = set()
        not_to_calculate: set[str] = set()

        def check_and_set(
            calc_state: LocalProjectCalculationState,
            attr: str,
            value: Any,
        ):
            if value is CALCULATE:
                to_calculate.add(attr)
            elif value is NOT_TO_CALCULATE:
                not_to_calculate.add(attr)
            elif value is not None and value != "":
                setattr(calc_state, attr, value)

        if local_project:
            state = LocalProjectCalculationState.from_local_project(local_project)
        else:
            state = LocalProjectCalculationState()
        check_and_set(state, "remote", remote)
        check_and_set(state, "git_repo", git_repo)
        check_and_set(state, "working_dir", working_dir)
        check_and_set(state, "ref", ref)
        check_and_set(state, "git_project", git_project)
        check_and_set(state, "git_service", git_service)
        check_and_set(state, "git_url", git_url)
        check_and_set(state, "full_name", full_name)
        check_and_set(state, "namespace", namespace)
        check_and_set(state, "repo_name", repo_name)
        self._refresh_the_state(state, to_calculate, not_to_calculate)
        # dataclasses.asdict cannot be used because some parts of the calculation state
        # are not deep-copyable (git.Repo and possibly ogr objects). A shallow copy suffices.
        state_dict = {
            field.name: getattr(state, field.name)
            for field in dataclasses.fields(state)
        }
        # Do not refresh the built local project (the logic will go away)
        return LocalProject(
            offline=self.offline,
            pr_id=pr_id,
            merge_pr=merge_pr,
            target_branch=target_branch,
            refresh=False,
            cache=self._cache,
            **state_dict,
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import itertools

from packit.dist_git_instance import DistGitInstance

DG_PR_COMMENT_KEY_SG_PR = "Source-git pull request ID"
DG_PR_COMMENT_KEY_SG_COMMIT = "Source-git commit"

# we store downstream content in source-git in this subdir
DISTRO_DIR = ".distro"
SRC_GIT_CONFIG = "source-git.yaml"
CONFIG_FILE_NAMES = {
    ".packit.yaml",
    ".packit.yml",
    "packit.yaml",
    "packit.yml",
}

# local branch name when checking out a PR before we merge it with the target branch
LP_TEMP_PR_CHECKOUT_NAME = "pr-changes"

DATETIME_FORMAT = "%Y%m%d%H%M%S%f"

# we create .tar.gz archives
DEFAULT_ARCHIVE_EXT = ".tar.gz"

DEFAULT_BODHI_NOTE = "New upstream release: {version}"
EXISTING_BODHI_UPDATE_REGEX = r".*Update for \S+ already exists.*"

FEDORA_DOMAIN = "fedoraproject.org"

DISTGIT_INSTANCES = {
    "fedpkg": DistGitInstance(
        hostname=f"src.{FEDORA_DOMAIN}",
        alternative_hostname=f"pkgs.{FEDORA_DOMAIN}",
        namespace="rpms",
    ),
    "fedpkg-stage": DistGitInstance(
        hostname=f"src.stg.{FEDORA_DOMAIN}",
        alternative_hostname=f"pkgs.stg.{FEDORA_DOMAIN}",
        namespace="rpms",
    ),
    "centpkg": DistGitInstance(
        hostname="gitlab.com",
        alternative_hostname=None,
        namespace="redhat/centos-stream/rpms",
    ),
    "centpkg-sig": DistGitInstance(
        hostname="gitlab.com",
        alternative_hostname=None,
        namespace="CentOS/{sig}/rpms",
    ),
}

DISTGIT_HOSTNAME_CANDIDATES = set(
    filter(
        lambda hostname: hostname is not None,
        itertools.chain.from_iterable(
            (distgit.hostname, distgit.alternative_hostname)
            for distgit in DISTGIT_INSTANCES.values()
        ),
    ),
)

COPR2GITHUB_STATE = {
    "running": ("pending", "The RPM build was triggered."),
    "pending": ("pending", "The RPM build has started."),
    "starting": ("pending", "The RPM build has started."),
    "importing": ("pending", "The RPM build has started."),
    "forked": ("pending", "The RPM build has started."),
    "waiting": ("pending", "The RPM build is in progress."),
    "succeeded": ("success", "RPMs were built successfully."),
    "failed": ("failure", "RPMs failed to be built."),
    "canceled": ("error", "The RPM build was canceled."),
    "skipped": ("error", "The build was already done."),
    "unknown": ("error", "Something went wrong."),
}

PACKIT_CONFIG_TEMPLATE = """# See the documentation for more information:
# https://packit.dev/docs/configuration/

specfile_path: {specfile_path}

# add or remove files that should be synced
files_to_sync:
    - {specfile_path}
    - .packit.yaml

# name in upstream package repository or registry (e.g. in PyPI)
upstream_package_name: {upstream_package_name}
# downstream (Fedora) RPM package name
downstream_package_name: {downstream_package_name}
"""

SANDCASTLE_WORK_DIR = "/sandcastle"
SANDCASTLE_IMAGE = "docker.io/usercont/sandcastle"
SANDCASTLE_DEFAULT_PROJECT = "myproject"
SANDCASTLE_PVC = "SANDCASTLE_PVC"

SYNCING_NOTE = (
    "This repository is maintained by packit.\nhttps://packit.dev/\n"
    "The file was generated using packit {packit_version}.\n"
)

PATCH_META_TRAILER_TOKENS = {
    "Patch-name",
    "Patch-id",
    "Patch-status",
    "Patch-present-in-specfile",
    "Ignore-patch",
    "No-prefix",
}

# shell code wrapped in lua as a python constant, what could go wrong?
# for real now: this horror is being used when creating source-git repos from upstream
# these macros override RPM default macros for setup and patching so that
# `rpmbuild -bp` leaves a git repo after invoking %prep
RPM_MACROS_FOR_PREP = [
    # we want both 'git init': here and in _packitpatch
    # if there are no patches, _packitpatch never gets invoked and
    # this will be invoked b/c of %autosetup
    "-D",
    "__scm_setup_patch(q) "
    "%{__git} init && "
    "%{__git} add -f . && "
    '%{__git} commit -q --allow-empty -a -m "%{NAME}-%{VERSION} base"',
    # %{1} = absolute path to the patch
    # %{2} = patch ID
    "-D",
    "__scm_apply_patch(qp:m:) "
    "%_packitpatch %{1} %{2} %{-p:-p%{-p*}} %{-q:-s} --fuzz=%{_default_patch_fuzz} "
    "%{_default_patch_flags}",
    # override patch program with _packitpatch
    "-D",
    "__patch %_packitpatch %{1} %{2} %{-p:-p%{-p*}} %{-q:-s} "
    "--fuzz=%{_default_patch_fuzz} %{_default_patch_flags}",
    "-D",
    "__scm_setup_git(q) "
    "%{__git} init %{-q} && "
    "%{__git} add -f . && "
    '%{__git} commit -q --allow-empty -a -m "%{NAME}-%{VERSION} base"',
    # commit_msg contains commit message of the last commit
    "-D",
    "__scm_apply_git_am(qp:m:) "
    "%{__git} am %{-q} %{-p:-p%{-p*}} && "
    "patch_name=`basename %{1}` && "
    "commit_msg=`%{__git} log --format=%B -n1` && "
    r'metadata_commit_msg=`printf "Patch-name: $patch_name"` && '
    '%{__git} commit --amend -m "$commit_msg" -m "$metadata_commit_msg"',
    # do the same of %autosetup -Sgit
    # that is, apply packit patch metadata to the patch commit
    # so packit can create the matching patch file
    "-D",
    "__scm_apply_git(qp:m:) "
    "%{__git} apply --index %{-p:-p%{-p*}} - && "
    "patch_name=`basename %{1}` && "
    r'metadata_commit_msg=`printf "Patch-name: $patch_name\\n"` && '
    '%{__git} commit %{-q} -m %{-m*} -m "${metadata_commit_msg}" --author "%{__scm_author}"',
]

COPR_SOURCE_SCRIPT = """
#!/bin/sh

git config --global user.email "hello@packit.dev"
git config --global user.name "Packit"
resultdir=$PWD
packit -d prepare-sources{package} --result-dir "$resultdir" {options}

"""

# Git-trailer tokens to mark the commit origin
FROM_DIST_GIT_TOKEN = "From-dist-git-commit"
FROM_SOURCE_GIT_TOKEN = "From-source-git-commit"

REPO_NOT_PRISTINE_HINT = (
    "Use 'git reset --hard HEAD' to reset changed files and "
    "'git clean -xdff' to delete untracked files and directories."
)

KOJI_BASEURL = "https://koji.fedoraproject.org/kojihub"

# key name -> default
CHROOT_SPECIFIC_COPR_CONFIGURATION = {
    "additional_packages": [],
    "additional_repos": [],
    # modules default to string because Copr stores it as string in the DB
    "additional_modules": "",
}

PACKAGE_LONG_OPTION = "--package"
PACKAGE_SHORT_OPTION = "-p"
PACKAGE_OPTION_HELP = (
    "Package to {action}, if more than one available, "
    "like in a monorepo configuration. "
    "Use it multiple times to select multiple packages."
    "Defaults to all the packages listed inside the config."
)

SYNC_RELEASE_DEFAULT_COMMIT_DESCRIPTION = (
    "{resolved_bugs}{upstream_tag}\n"
    "{upstream_commit_info}\n"
    "\n"
    "Commit authored by Packit automation (https://packit.dev/)\n"
)

SYNC_RELEASE_PR_DESCRIPTION = (
    "{upstream_tag_info}\n"
    "{upstream_commit_info}\n"
    "{release_monitoring_info}"
    "{resolved_bugzillas_info}"
)

SYNC_RELEASE_PR_INSTRUCTIONS = (
    "\n---\n\n"
    "If you need to do any change in this pull request, you can clone Packit's fork "
    "and push directly to the source branch of this PR (provided you have commit access "
    "to this repository):\n"
    "```\n"
    "git clone ssh://$YOUR_USER@pkgs.fedoraproject.org/forks/{user}/rpms/{package}.git\n"
    "cd {package}\n"
    "git checkout {branch}\n"
    "git push origin {branch}\n"
    "```\n"
    "\n---\n\n"
    "Alternatively, if you already have the package repository cloned, "
    "you can just fetch the Packit's fork:\n"
    "```\n"
    "cd {package}\n"
    "git remote add packit ssh://$YOUR_USER@pkgs.fedoraproject.org/forks/{user}/rpms/{package}.git\n"
    "git fetch packit refs/heads/{branch}\n"
    "git checkout {branch}\n"
    "git push packit {branch}\n"
    "```\n"
    "If you have the `koji_build` job configured as well, make sure to configure "
    "the `allowed_pr_authors` and/or `allowed_committers` (see [the docs]"
    "(https://packit.dev/docs/configuration/downstream/koji_build#optional-parameters)) "
    "since by default, Packit reacts only to its own PRs.\n"
    "\n---\n\n"
    "Before pushing builds/updates, please remember to check the new "
    "version against the "
    "[packaging guidelines](https://docs.fedoraproject.org/en-US/packaging-guidelines/).\n\n"
    "For example, please:\n\n"
    "- check that the new sources only contain permissible content\n"
    "- check that the license of the new version has not changed\n"
    "- check for any API/ABI and other changes that may break dependent packages\n"
    "- check the autogenerated changelog\n"
)

COMMIT_ACTION_DIVIDER = "---%<--- snip ---%<--- here ---%<---\n"
RELEASE_MONITORING_PROJECT_URL = "https://release-monitoring.org/project/{project_id}"
BUGZILLA_URL = "https://bugzilla.redhat.com/show_bug.cgi?id={bug_id}"
BUGZILLA_HOSTNAME = "bugzilla.redhat.com"
ANITYA_MONITORING_CHECK_URL = (
    "https://src.fedoraproject.org/_dg/anitya/rpms/{package_name}"
)
RELEASE_MONITORING_PACKAGE_CHECK_URL = "https://release-monitoring.org//api/v2/packages/?name={package_name}&distribution=Fedora"

# connection timeout and read timeout in seconds
# with connection timeout, the actual value will usually be a multiple of what is configured,
# depending on the number of IP addresses for the target domain
HTTP_REQUEST_TIMEOUT = (10, 30)

FAST_FORWARD_MERGE_INTO_KEY = "fast_forward_merge_into"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from typing import Optional

import requests

from packit.constants import HTTP_REQUEST_TIMEOUT

logger = logging.getLogger(__name__)


def get_upstream_version(package_name: str) -> Optional[str]:
    """
    Gets the latest upstream version of the specified package.

    Args:
        package_name: Package name (name of SRPM in Fedora).

    Returns:
        The latest upstream version or None if no matching project
        was found on release-monitoring.org.
    """

    def query(endpoint, **kwargs):
        try:
            response = requests.get(
                f"https://release-monitoring.org/api/{endpoint}",
                params=kwargs,
                timeout=HTTP_REQUEST_TIMEOUT,
            )
        except requests.exceptions.RequestException as e:
            logger.debug(f"release-monitoring.org query failed: {e!s}")
            return {}
        if not response.ok:
            return {}
        return response.json()

    if not package_name:
        return None
    result = query(f"project/Fedora/{package_name}")
    version = result.get("version")
    if not version:
        # if there is no Fedora mapping, try using package name as project name
        result = query("projects", pattern=package_name)
        projects = result.get("projects", [])
        project: dict = next(
            iter(p for p in projects if p.get("name") == package_name),
            {},
        )
        version = project.get("version")
    return version



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import logging
import os
from pathlib import Path

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.schema import JobConfigSchema
from packit.utils.changelog_helper import ChangelogHelper

logger = logging.getLogger("packit")


def load_job_config(job_config):
    if job_config:
        try:
            return JobConfigSchema().loads(job_config)
        except Exception as ex:
            logger.error(f"Loading of JobConfig wasn't successful: {ex}")
    return None


@click.command("prepare-sources", context_settings=get_context_settings())
@click.option(
    "--result-dir",
    metavar="DIR",
    help="Copy the sources into DIR. By default, `prepare_sources_result` directory "
    "in the current working directory is created.",
)
@click.option(
    "--upstream-ref",
    default=None,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@click.option(
    "--merged-ref",
    default=None,
    help="Git ref used to identify correct most recent tag.",
)
@click.option(
    "--update-release/--no-update-release",
    default=None,
    help=(
        "Specifies whether to update Release. "
        "Defaults to value set in configuration, which defaults to yes."
    ),
)
@click.option(
    "--bump/--no-bump",
    default=None,
    help="Deprecated. Use --[no-]update-release instead.",
)
@click.option(
    "--release-suffix",
    default=None,
    type=click.STRING,
    help="Specifies release suffix. Allows to override default generated:"
    "{current_time}.{sanitized_current_branch}{git_desc_suffix}",
)
@click.option(
    "--default-release-suffix",
    default=False,
    is_flag=True,
    help=(
        "Allows to use default, packit-generated, release suffix when some "
        "release_suffix is specified in the configuration."
    ),
)
@click.option(
    "--job-config-index",
    default=None,
    type=click.INT,
    help="Internal option to override package config found in the repository "
    "with job config with given index (needed for packit service).",
)
@click.option(
    "--ref",
    default="",
    type=click.STRING,
    help="Git reference to checkout.",
)
@click.option(
    "--pr-id",
    default=None,
    type=click.STRING,
    help="Specifies PR to checkout.",
)
@click.option(
    "--merge-pr/--no-merge-pr",
    is_flag=True,
    default=True,
    help="Specifies whether to merge PR into the base branch in case pr-id is specified.",
)
@click.option(
    "--target-branch",
    default=None,
    type=click.STRING,
    help="Specifies target branch which PR should be merged into.",
)
@click.option(
    "--create-symlinks/--no-create-symlinks",
    is_flag=True,
    default=True,
    help="Specifies whether Packit should create symlinks or copy the "
    "files (e.g. archive outside specfile dir).",
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="prepare"),
)
@click.argument(
    "path_or_url",
    type=LocalProjectParameter(
        ref_param_name="ref",
        pr_id_param_name="pr_id",
        merge_pr_param_name="merge_pr",
        target_branch_param_name="target_branch",
    ),
    default=os.path.curdir,
)
@pass_config
@cover_packit_exception
@iterate_packages
def prepare_sources(
    config,
    path_or_url,
    job_config_index,
    upstream_ref,
    merged_ref,
    update_release,
    bump,
    release_suffix,
    default_release_suffix,
    result_dir,
    ref,
    pr_id,
    merge_pr,
    target_branch,
    create_symlinks,
    package_config,
):
    """
    Prepare sources for a new SRPM build using content of the upstream repository.
    Determine version, create an archive or download upstream and create patches for sourcegit,
    fix/update the specfile to use the right archive, download the remote sources.
    Behaviour can be customized by specifying actions (post-upstream-clone, get-current-version,
    create-archive, create-patches, fix-spec-file) in the configuration.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """

    if not result_dir:
        path = package_config.paths[0].strip("./\\")
        result_dir = (
            Path.cwd().joinpath(f"prepare_sources_result_{path}")
            if path
            else Path.cwd().joinpath("prepare_sources_result")
        )
        logger.debug(f"Setting result_dir to: {result_dir}")
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
        job_config_index=job_config_index,
    )
    if bump is not None:
        if update_release is not None:
            raise click.UsageError(
                "--[no-]bump and --[no-]update-release are mutually exclusive",
            )
        logger.warning("--[no-]bump is deprecated. Use --[no-]update-release instead.")
        update_release = bump
    release_suffix = ChangelogHelper.resolve_release_suffix(
        api.package_config,
        release_suffix,
        default_release_suffix,
    )

    api.prepare_sources(
        upstream_ref=upstream_ref,
        merged_ref=merged_ref,
        update_release=update_release,
        release_suffix=release_suffix,
        result_dir=result_dir,
        create_symlinks=create_symlinks,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock
from munch import Munch


@pytest.fixture()
def query_response():
    return Munch(
        {
            "updates": [
                Munch(
                    {
                        "autokarma": True,
                        "autotime": True,
                        "stable_karma": 3,
                        "stable_days": 7,
                        "unstable_karma": -3,
                        "requirements": "",
                        "require_bugs": True,
                        "require_testcases": True,
                        "display_name": "",
                        "notes": "New upstream release: 0.5.1",
                        "type": "enhancement",
                        "status": "testing",
                        "request": None,
                        "severity": "unspecified",
                        "suggest": "unspecified",
                        "locked": False,
                        "pushed": True,
                        "critpath": False,
                        "close_bugs": True,
                        "date_submitted": "2019-08-26 10:53:13",
                        "date_modified": None,
                        "date_approved": None,
                        "date_pushed": "2019-08-26 18:21:12",
                        "date_testing": "2019-08-26 18:21:12",
                        "date_stable": None,
                        "alias": "FEDORA-2019-89c99f680c",
                        "test_gating_status": "ignored",
                        "meets_testing_requirements": False,
                        "url": "https://bodhi.fedoraproject.org/updates/FEDORA-2019-89c99f680c",
                        "title": "packit-0.5.1-1.fc30",
                        "release": Munch(
                            {
                                "name": "F30",
                                "long_name": "Fedora 30",
                                "version": "30",
                                "id_prefix": "FEDORA",
                                "branch": "f30",
                                "dist_tag": "f30",
                                "stable_tag": "f30-updates",
                                "testing_tag": "f30-updates-testing",
                                "candidate_tag": "f30-updates-candidate",
                                "pending_signing_tag": "f30-signing-pending",
                                "pending_testing_tag": "f30-updates-testing-pending",
                                "pending_stable_tag": "f30-updates-pending",
                                "override_tag": "f30-override",
                                "mail_template": "fedora_errata_template",
                                "state": "current",
                                "composed_by_bodhi": True,
                                "create_automatic_updates": None,
                                "package_manager": "unspecified",
                                "testing_repository": None,
                                "composes": [],
                            },
                        ),
                        "comments": [
                            Munch(
                                {
                                    "id": 1014223,
                                    "karma": 0,
                                    "karma_critpath": 0,
                                    "text": "This update has been submitted for testing by ttomece",
                                    "timestamp": "2019-08-26 10:53:13",
                                    "update_id": 149516,
                                    "user_id": 91,
                                    "bug_feedback": [],
                                    "testcase_feedback": [],
                                    "user": Munch(
                                        {
                                            "id": 91,
                                            "name": "bodhi",
                                            "email": None,
                                            "avatar": None,
                                            "openid": "bodhi.id.fedoraproject.org",
                                            "groups": [],
                                        },
                                    ),
                                },
                            ),
                            Munch(
                                {
                                    "id": 1014224,
                                    "karma": 0,
                                    "karma_critpath": 0,
                                    "text": "This update's test gating status has been changed.",
                                    "timestamp": "2019-08-26 10:53:13",
                                    "update_id": 149516,
                                    "user_id": 91,
                                    "bug_feedback": [],
                                    "testcase_feedback": [],
                                    "user": Munch(
                                        {
                                            "id": 91,
                                            "name": "bodhi",
                                            "email": None,
                                            "avatar": None,
                                            "openid": "bodhi.id.fedoraproject.org",
                                            "groups": [],
                                        },
                                    ),
                                },
                            ),
                            Munch(
                                {
                                    "id": 1014225,
                                    "karma": 0,
                                    "karma_critpath": 0,
                                    "text": "This update's test gating status has been changed.",
                                    "timestamp": "2019-08-26 10:53:16",
                                    "update_id": 149516,
                                    "user_id": 91,
                                    "bug_feedback": [],
                                    "testcase_feedback": [],
                                    "user": Munch(
                                        {
                                            "id": 91,
                                            "name": "bodhi",
                                            "email": None,
                                            "avatar": None,
                                            "openid": "bodhi.id.fedoraproject.org",
                                            "groups": [],
                                        },
                                    ),
                                },
                            ),
                            Munch(
                                {
                                    "id": 1016059,
                                    "karma": 0,
                                    "karma_critpath": 0,
                                    "text": "This update has been pushed to testing.",
                                    "timestamp": "2019-08-27 18:22:32",
                                    "update_id": 149516,
                                    "user_id": 91,
                                    "bug_feedback": [],
                                    "testcase_feedback": [],
                                    "user": Munch(
                                        {
                                            "id": 91,
                                            "name": "bodhi",
                                            "email": None,
                                            "avatar": None,
                                            "openid": "bodhi.id.fedoraproject.org",
                                            "groups": [],
                                        },
                                    ),
                                },
                            ),
                        ],
                        "builds": [
                            Munch(
                                {
                                    "nvr": "packit-0.5.1-1.fc30",
                                    "release_id": 28,
                                    "signed": True,
                                    "type": "rpm",
                                    "epoch": 0,
                                },
                            ),
                        ],
                        "compose": None,
                        "bugs": [],
                        "user": Munch(
                            {
                                "id": 754,
                                "name": "ttomecek",
                                "email": "ttomecek@redhat.com",
                                "avatar": None,
                                "openid": "ttomecek.id.fedoraproject.org",
                                "groups": [
                                    Munch({"name": "provenpackager"}),
                                    Munch({"name": "packager"}),
                                ],
                            },
                        ),
                        "updateid": "FEDORA-2019-89c99f680c",
                        "karma": 0,
                        "content_type": "rpm",
                        "test_cases": [],
                    },
                ),
            ],
            "page": 1,
            "pages": 1,
        },
    )


@pytest.fixture()
def request_response():
    return Munch(
        {
            "update": Munch(
                {
                    "autokarma": True,
                    "autotime": True,
                    "stable_karma": 3,
                    "stable_days": 7,
                    "unstable_karma": -3,
                    "requirements": "",
                    "require_bugs": True,
                    "require_testcases": True,
                    "display_name": "",
                    "notes": "New upstream release: 0.5.1",
                    "type": "enhancement",
                    "status": "stable",
                    "request": None,
                    "severity": "unspecified",
                    "suggest": "unspecified",
                    "locked": False,
                    "pushed": True,
                    "critpath": False,
                    "close_bugs": True,
                    "date_submitted": "2019-08-26 10:53:13",
                    "date_modified": None,
                    "date_approved": None,
                    "date_pushed": "2019-08-26 18:21:12",
                    "date_testing": "2019-08-26 18:21:12",
                    "date_stable": "2019-09-03 08:21:12",
                    "alias": "FEDORA-2019-89c99f680c",
                    "test_gating_status": "ignored",
                    "meets_testing_requirements": False,
                    "url": "https://bodhi.fedoraproject.org/updates/FEDORA-2019-89c99f680c",
                    "title": "packit-0.5.1-1.fc30",
                    "release": Munch(
                        {
                            "name": "F30",
                            "long_name": "Fedora 30",
                            "version": "30",
                            "id_prefix": "FEDORA",
                            "branch": "f30",
                            "dist_tag": "f30",
                            "stable_tag": "f30-updates",
                            "testing_tag": "f30-updates-testing",
                            "candidate_tag": "f30-updates-candidate",
                            "pending_signing_tag": "f30-signing-pending",
                            "pending_testing_tag": "f30-updates-testing-pending",
                            "pending_stable_tag": "f30-updates-pending",
                            "override_tag": "f30-override",
                            "mail_template": "fedora_errata_template",
                            "state": "current",
                            "composed_by_bodhi": True,
                            "create_automatic_updates": None,
                            "package_manager": "unspecified",
                            "testing_repository": None,
                            "composes": [],
                        },
                    ),
                    "comments": [
                        Munch(
                            {
                                "id": 1014223,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted for testing by ttomecek. ",
                                "timestamp": "2019-08-26 10:53:13",
                                "update_id": 149516,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "avatar": None,
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    },
                                ),
                            },
                        ),
                        Munch(
                            {
                                "id": 1014224,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update's test gating status has been changed.",
                                "timestamp": "2019-08-26 10:53:13",
                                "update_id": 149516,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "avatar": None,
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    },
                                ),
                            },
                        ),
                        Munch(
                            {
                                "id": 1014225,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update's test gating status has been changed.",
                                "timestamp": "2019-08-26 10:53:16",
                                "update_id": 149516,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "avatar": None,
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    },
                                ),
                            },
                        ),
                        Munch(
                            {
                                "id": 1016059,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to testing.",
                                "timestamp": "2019-08-27 18:22:32",
                                "update_id": 149516,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "avatar": None,
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    },
                                ),
                            },
                        ),
                    ],
                    "builds": [
                        Munch(
                            {
                                "nvr": "packit-0.5.1-1.fc30",
                                "release_id": 28,
                                "signed": True,
                                "type": "rpm",
                                "epoch": 0,
                            },
                        ),
                    ],
                    "compose": None,
                    "bugs": [],
                    "user": Munch(
                        {
                            "id": 754,
                            "name": "ttomecek",
                            "email": "ttomecek@redhat.com",
                            "avatar": None,
                            "openid": "ttomecek.id.fedoraproject.org",
                            "groups": [
                                Munch({"name": "provenpackager"}),
                                Munch({"name": "packager"}),
                            ],
                        },
                    ),
                    "updateid": "FEDORA-2019-89c99f680c",
                    "karma": 0,
                    "content_type": "rpm",
                    "test_cases": [],
                },
            ),
        },
    )


def test_push_updates(
    cwd_upstream_or_distgit,
    api_instance,
    query_response,
    request_response,
):
    from bodhi.client.bindings import BodhiClient

    u, d, api = api_instance

    flexmock(BodhiClient)
    BodhiClient.should_receive("query").and_return(query_response).once()
    BodhiClient.should_receive("request").with_args(
        update="FEDORA-2019-89c99f680c",
        request="stable",
    ).and_return(request_response).once()

    flexmock(
        BodhiClient,
        ensure_auth=lambda: None,  # this is where the browser/OIDC fun happens
    )

    api.push_updates()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Functional tests the validate-config command
"""

from packit.utils.commands import cwd
from tests.functional.spellbook import call_real_packit_and_return_exit_code


def test_srpm_command_for_path(upstream_or_distgit_path, tmp_path):
    with cwd(tmp_path):
        call_real_packit_and_return_exit_code(
            parameters=["--debug", "validate-config", str(upstream_or_distgit_path)],
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from pathlib import Path

from flexmock import flexmock
from requre.cassette import DataTypes
from requre.online_replacing import (
    record_requests_for_all_methods,
)
from specfile import Specfile

from packit.base_git import PackitRepositoryBase
from packit.config import CommonPackageConfig, PackageConfig
from packit.config.sources import SourcesItem
from tests_recording.testbase import PackitTest


@record_requests_for_all_methods()
class ProposeUpdate(PackitTest):
    def cassette_setup(self, cassette):
        cassette.data_miner.data_type = DataTypes.Dict

    def test_download_remote_sources_via_spec(self):
        """
        Use case: package_config.sources and Source0 are out of sync,
        make sure packit downloads correct archive specifiec in the spec file
        """
        # we should use an actual git.centos.org url but we'd store the tarball in our history
        # which we don't want I'd say
        # "https://git.centos.org/sources/rsync/c8s/82e7829c0b3cefbd33c233005341e2073c425629"
        git_centos_org_url = "https://example.org/"
        package_config = PackageConfig(
            packages={
                "rsync": CommonPackageConfig(
                    downstream_package_name="rsync",
                    specfile_path="rsync.spec",
                    sources=[
                        SourcesItem(
                            path="rsync-3.1.2.tar.gz",
                            url=git_centos_org_url,
                        ),
                    ],
                ),
            },
            jobs=[],
        )
        # same drill here, let's not store tarballs in our git-history
        # source = "https://download.samba.org/pub/rsync/src/rsync-3.1.3.tar.gz"
        source = "https://httpbin.org/anything/rsync-3.1.3.tar.gz"

        base_git = PackitRepositoryBase(
            config=flexmock(),
            package_config=package_config,
        )
        specfile_content = (
            "Name: rsync\n"
            "Version: 3.1.3\n"
            "Release: 1\n"
            f"Source0: {source}\n"
            "License: GPLv3+\n"
            "Summary: rsync\n"
            "%description\nrsync\n"
        )
        tmp = Path(self.static_tmp)
        spec_path = tmp / "rsync.spec"
        spec_path.write_text(specfile_content)
        specfile = Specfile(spec_path, sourcedir=tmp, autosave=True)
        flexmock(base_git).should_receive("specfile").and_return(specfile)

        def mocked_is_file():
            import inspect

            # return False only if Path.is_file() is called directly from within
            # the download_remote_sources() method
            # this is necessary because specfile relies on Path.is_file() as well
            return inspect.stack()[3].function != "download_remote_sources"

        flexmock(Path).should_receive("is_file").replace_with(mocked_is_file)

        base_git.download_remote_sources()

        expected_path = tmp / "rsync-3.1.3.tar.gz"
        assert Path(expected_path).exists()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import re
import subprocess
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

import git
import yaml
from git.exc import GitCommandError
from ogr.parsing import RepoUrl, parse_git_repo

from packit.constants import COMMIT_ACTION_DIVIDER
from packit.exceptions import PackitException

logger = logging.getLogger(__name__)


class RepositoryCache:
    """
    Cache for git repositories base on the reference option of `git clone`.

    * The cache is located in the specified directory
      and contains separate git repository for each project.
    * Project name is used to match the git project in the cache.
    """

    def __init__(self, cache_path: Union[str, Path], add_new=False) -> None:
        self.cache_path = (
            Path(cache_path) if isinstance(cache_path, str) else cache_path
        )
        self.add_new = add_new
        logger.debug(
            f"Instantiation of the repository cache at {self.cache_path}. "
            f"New projects will {'not ' if not self.add_new else ''}be added.",
        )
        self.projects_added: list[str] = []
        self.projects_cloned_using_cache: list[str] = []

    @property
    def cached_projects(self) -> list[str]:
        """Project names we have in the cache."""
        if not self.cache_path.is_dir():
            self.cache_path.mkdir(parents=True)
        return [f.name for f in self.cache_path.iterdir() if f.is_dir()]

    def _clone(self, **kwargs) -> git.Repo:
        """Wrapper around git function so we are able to check the call in tests more easily."""
        return git.repo.Repo.clone_from(**kwargs)

    def get_repo(
        self,
        url: str,
        directory: Union[Path, str, None] = None,
    ) -> git.Repo:
        """
        Clone the repository.
        * If we have this repository in a cache, use the cached repo as a reference when cloning.
        * If we don't have this repository in a cache and {add_new} is True,
          clone the repository to cache first and then use it as a reference.

        :param url: will be used to clone the repo
        :param directory: target path for cloning the repository
        :return: cloned repository
        """
        directory = str(directory) if directory else tempfile.mkdtemp()

        if is_git_repo(directory=directory):
            logger.debug(f"Repo already exists in {directory}.")
            return git.repo.Repo(directory)

        logger.debug(
            f"Cloning repo {url} -> {directory} using repository cache at {self.cache_path}",
        )
        cached_projects = self.cached_projects
        cached_projects_str = "\n".join(f"- {project}" for project in cached_projects)
        logger.debug(
            f"Repositories in the cache ({len(cached_projects)} "
            f"project(s)):\n{cached_projects_str}",
        )

        project_name = RepoUrl.parse(url).repo
        reference_repo = self.cache_path.joinpath(project_name)
        if project_name not in cached_projects and self.add_new:
            logger.debug(f"Creating reference repo: {reference_repo}")
            self._clone(url=url, to_path=str(reference_repo), tags=True)
            self.projects_added.append(project_name)

        if self.add_new or project_name in cached_projects:
            logger.debug(f"Using reference repo: {reference_repo}")
            self.projects_cloned_using_cache.append(project_name)
            return self._clone(
                url=url,
                to_path=directory,
                tags=True,
                reference=str(reference_repo),
            )

        return self._clone(url=url, to_path=directory, tags=True)


def is_git_repo(directory: Union[Path, str]) -> bool:
    """
    Test, if the directory is a git repo.
    (Has .git subdirectory or 'gitdir' file?)
    """
    return Path(directory, ".git").exists()


def get_repo(url: str, directory: Optional[Union[Path, str]] = None) -> git.Repo:
    """
    Use directory as a git repo or clone repo to the tempdir.
    """
    directory = str(directory) if directory else tempfile.mkdtemp()

    if is_git_repo(directory=directory):
        logger.debug(f"Repo already exists in {directory}.")
        return git.repo.Repo(directory)
    logger.info(f"Cloning repo {url} -> {directory}")
    return git.repo.Repo.clone_from(url=url, to_path=directory, tags=True)


def get_namespace_and_repo_name(url: str) -> tuple[Optional[str], str]:
    parsed_git_repo = parse_git_repo(url)
    if parsed_git_repo is None or not parsed_git_repo.repo:
        raise PackitException(
            f"Invalid URL format, can't obtain namespace and repository name: {url}",
        )
    return parsed_git_repo.namespace, parsed_git_repo.repo


def is_a_git_ref(repo: git.Repo, ref: str) -> bool:
    try:
        commit = repo.commit(ref)
        return bool(commit)
    except git.BadName:
        return False


def get_default_branch(repository: git.Repo) -> str:
    """
    Returns default branch for newly created repos in the parent directory of
    passed in repository. Accepts `repository` to ensure the closest override of
    git configuration is used.

    Args:
        repository (git.Repo): Git repository closest to the directory where
            the configuration is applied.

    Returns:
        Default branch for new repos, if not supported or not configured returns
        `master`.
    """
    config = repository.config_reader()
    return config.get_value("init", "defaultBranch", "master")


def git_remote_url_to_https_url(
    inp: str,
    with_dot_git_suffix: Optional[bool] = True,
) -> str:
    """
    turn provided git remote URL to https URL:
    returns empty string if the input can't be processed
    """
    logger.debug(f"Parsing git remote URL {inp!r} and converting it to https-like URL.")
    parsed_repo = parse_git_repo(inp)
    if not parsed_repo or not parsed_repo.hostname:
        logger.debug(f"{inp!r} is not an URL we recognize.")
        return ""

    if inp.startswith(("http", "https")):
        logger.debug(f"Provided input {inp!r} is an url.")
        if inp.endswith(".git"):
            return inp if with_dot_git_suffix else inp[:-4]

        return inp

    optional_suffix = ".git" if inp.endswith(".git") and with_dot_git_suffix else ""
    url_str = "https://{}/{}/{}{}".format(
        parsed_repo.hostname,
        parsed_repo.namespace,
        parsed_repo.repo,
        optional_suffix,
    )

    logger.debug(f"URL {inp!r} turned into HTTPS {url_str!r}")
    return url_str


def get_current_version_command(
    glob_pattern: str,
    refs: Optional[str] = "tags",
) -> list[str]:
    """
    Returns command that find latest git reference matching given pattern.

    :param glob_pattern: pattern that is used to find latest ref
    :param refs: specifies what kind of ref is used; \
        default is `"tags"` that searches through all tags (including non-annotated), \
        pass `None` to search only annotated tags or `"all"` to search through \
        all refs (including branches and remote refs)
    :return: command to find latest ref
    """
    return [
        "git",
        "describe",
        "--abbrev=0",
        f"--{refs}" if refs else "",
        "--match",
        glob_pattern,
    ]


def create_new_repo(cwd: Path, switches: list[str]):
    subprocess.check_call(["git", "init", *switches, str(cwd)])
    # TODO: Replace with -b / --initial-branch in `git init` when possible
    if "--bare" not in switches:
        subprocess.check_call(["git", "checkout", "-b", "main"], cwd=cwd)
    else:
        subprocess.check_call(
            ["git", "symbolic-ref", "HEAD", "refs/heads/main"],
            cwd=cwd,
        )


def git_patch_ish(patch: str) -> str:
    """
    Massage patch to look like a Git-style patch, so that it can
    be passed to 'git patch-id' in order to calculate a patch-id.

    :param patch: Patch to transform.
    :return: Transformed patch.
    """
    # Prettend as if format is 'diff --git'
    pattern = re.compile(r"^diff -\w+ ", flags=re.MULTILINE)
    repl = r"diff --git "
    patch = re.sub(pattern, repl, patch)

    # Remove timestamps from comparison lines
    pattern = re.compile(r"^((---|\+\+\+) .+)\t\d{4}.+$", flags=re.MULTILINE)
    repl = r"\1"
    patch = re.sub(pattern, repl, patch)

    # Add missing 'diff --git' lines
    if "diff --git " not in patch:
        # Timestamps (see above) already need to be removed
        # for this substitution pattern to work.
        pattern = re.compile(r"(\n--- (.+)\n\+\+\+ (.+)\n)")
        repl = r"\ndiff --git \2 \3\1"
        patch = re.sub(pattern, repl, patch)

    return patch


def get_message_from_metadata(metadata: dict, header: Optional[str] = None) -> str:
    if not isinstance(metadata, dict):
        raise PackitException(
            f"We can save only dictionaries to metadata. Not {metadata}",
        )

    content = (
        yaml.dump(metadata, indent=4, default_flow_style=False) if metadata else ""
    )
    if not header:
        return content

    return f"{header}\n\n{content}"


def get_metadata_from_message(commit: git.Commit) -> Optional[dict]:
    """
    Tries to load yaml format from the git message.

    We are skipping first line until
    the rest of the content is yaml-loaded to dictionary (yaml object type).

    If nothing found, we return None.

    Reference:
    https://gitpython.readthedocs.io/en/stable/reference.html
    ?highlight=archive#module-git.objects.commit

    e.g.:

    I)
    key: value
    another: value
    -> {"key": "value", "another": "value"}

    II)
    On sentence.

    key: value
    another: value
    -> {"key": "value", "another": "value"}

    III)
    A lot of
    text

    before keys.

    key: value
    another: value
    -> {"key": "value", "another": "value"}

    IV)
    Other values are supported as well:

    key:
    - first
    - second
    - third

    :param commit: git.Commit object
    :return: dict loaded from message if it satisfies the rules above
    """
    splitted_message = commit.message.split("\n")

    for i in range(len(splitted_message)):
        message_part = "\n".join(splitted_message[i:])
        try:
            loaded_part = yaml.safe_load(message_part)
        except yaml.YAMLError:
            continue

        if isinstance(loaded_part, dict):
            return loaded_part

    return None


def shorten_commit_hash(commit_hash: str) -> str:
    """
    Shortens commit hash to first 8 characters.

    Args:
        commit_hash: Commit hash to be shortened.

    Returns:
        First 8 characters of the commit hash.
    """
    return commit_hash[:8]


def get_next_commit(repo: git.Repo, commit: str) -> Optional[str]:
    """Returns the commit following the specified commit.

    Args:
        repo: Git repo to search the commit in.
        commit: Hash of the commit to get the next commit of.

    Returns:
        The hash of the commit following the given commit or None if no
        such commit exists, i.e. the given commit is the HEAD.
    """
    commits = list(repo.iter_commits(f"{commit}..", ancestry_path=True))
    return commits[-1].hexsha if commits else None


def commit_exists(repo: git.Repo, commit: str) -> bool:
    """Checks whether a commit with the given hash exists.

    Args:
        repo: Git repo to check if the commit exists in.
        commit: Hash of the commit to check.

    Returns:
        Whether a commit with such hash exists
    """
    try:
        list(repo.iter_commits(f"{commit}..{commit}"))
    except GitCommandError:
        return False
    else:
        return True


def get_commit_diff(commit: git.Commit) -> list[git.Diff]:
    """Get modified files of the given commit.

    Args:
        commit: Commit to get the diff of.

    Returns:
        List of git.Diff containing information about the modified files
        in the given commit.
    """
    if len(commit.parents) == 1:
        return commit.parents[0].diff(commit, create_patch=True)
    if len(commit.parents) == 0:
        # First commit in the repo
        return commit.diff(git.NULL_TREE, create_patch=True)
    # Probably a merge commit, we can't do much about it
    return []


def get_commit_hunks(repo: git.Repo, commit: git.Commit) -> list[str]:
    """Get a list of hunks of the given commit.

    Args:
        repo: Git repo which the commit belongs to.
        commit: Commit to get hunks of.

    Returns:
        List of split commit hunks stored as strings where each string
        represents changes of a single file.
    """
    patch = repo.git.show(commit, format="", color="never")
    hunk_start_re = re.compile(r"diff --git a/.+ b/.+")
    section_start = 0
    result = []
    patch_lines = patch.splitlines()
    for i, line in enumerate(patch_lines):
        if hunk_start_re.match(line):
            section = patch_lines[section_start:i]
            if section:
                result.append("\n".join(section))
            section_start = i
    # The last section
    section = patch_lines[section_start:]
    if section:
        result.append("\n".join(section))
    return result


def is_the_repo_pristine(repo: git.Repo) -> bool:
    """Checks whether the repository is pristine.

    Args:
        repo: Git repo to check.

    Returns:
        Whether the repo is pristine.
    """
    return not repo.git.diff() and not repo.git.clean("-xdn")


def get_file_author(repo: git.Repo, filename: str) -> str:
    """Get the original author of 'filename' in 'repo'

    Args:
        repo: Git-repo where the file is commited.
        filename: Name of the file.

    Returns:
        The original (first) author of the file in the
        "A U Thor <author@example.com>" format.
    """
    author, author_mail = "", ""
    for line in repo.git.blame(filename, line_porcelain=True).splitlines():
        token, _, value = line.partition(" ")
        if token == "author":
            author = value
        elif token == "author-mail":
            author_mail = value
        elif token == "filename":
            # End of the first blame-block
            break
    return f"{author} {author_mail}"


@contextmanager
def commit_message_file(
    subject: str,
    message: Optional[str] = None,
    trailers: Optional[list[tuple[str, str]]] = None,
) -> Generator[str, None, None]:
    """Context manager to yield a commit message file

    Which then can be used by a commit operation.

    This handles a few things:
    - It concatenates a commit message subject line and a message.
      Though: if 'subject' is the complete commit message that'll also work.
    - Adds the Git-trailers specified by 'trailers'.

    Args:
        subject: Subject line of the commit message.
        message: Message following the subject line.
        trailers: List Git-trailers to be added to the commit message.

    Yields:
        Name of the temporary file with the prepared commit message.
    """
    with tempfile.NamedTemporaryFile(mode="w+t") as fp:
        fp.writelines([subject, "\n"])
        if message:
            fp.writelines(["\n", message, "\n"])
        fp.seek(0)
        if trailers:
            args = ["--in-place", "--if-exists", "replace"]
            for token, value in trailers:
                args += ["--trailer", f"{token}={value}"]
            # 'interpret-trailers' doesn't require a working directory,
            # so a plain Git command can be used here.
            git.cmd.Git().interpret_trailers(*args, fp.name)
        yield fp.name


def get_commit_message_from_action(
    output: Optional[list[str]],
    default_title: str,
    default_description: str,
) -> tuple[str, str]:
    """
    Parse the output of the commit action and in case the action is not defined,
    no output has been produced or it couldn't be parsed, return the defaults.

    Args:
        output: Output produced by the `commit-message` action.
        default_title: Commit title that is used in case the produced commit
            message hasn't been produced, is malformed or couldn't be parsed.
        default_description: Commit description that is used in case the
            produced commit message hasn't been produced, is malformed or
            couldn't be parsed.

    Returns:
        Pair of the commit title and commit message that will be used.
    """
    # no output has been produced, or action doesn't exist
    if not output:
        return (default_title, default_description)

    # split by the divider
    split_output = "".join(output).rsplit(sep=COMMIT_ACTION_DIVIDER, maxsplit=1)

    # -1 ensures we're taking just the message, ignoring, if any, debugging
    # output
    whole_message = split_output[-1]

    # we split only once, cause we separate the commit title
    split_commit_message = whole_message.split("\n\n", maxsplit=1)

    # nothing found or empty title
    if len(split_commit_message) < 1 or not split_commit_message[0]:
        return default_title, default_description

    title = split_commit_message[0]
    description = split_commit_message[1] if len(split_commit_message) > 1 else ""

    # it is later reconstructed in a generic way for both defaults and override,
    # so we don't care about the whitespace at the beginning and the end
    return title.strip(), description.strip()


def get_tag_link(git_url: str, upstream_tag: str) -> str:
    """
    Get link to the tag of a Git repo.
    """
    link = ""
    git_repo = parse_git_repo(git_url)
    if not git_repo:
        return ""

    forge = git_repo.hostname
    if not forge:
        return ""

    if forge == "github.com":
        link = f"{git_url}/releases/tag/{upstream_tag}"
    # GitLab or GitLab instances (e.g. gitlab.gnome.org)
    elif "gitlab" in forge:
        link = f"{git_url}/-/tags/{upstream_tag}"

    return link


def get_commit_link(git_url: str, upstream_commit: str) -> str:
    """
    Get link to the commit of a Git repo.
    """
    link = ""
    git_repo = parse_git_repo(git_url)
    if not git_repo:
        return ""

    forge = git_repo.hostname
    if not forge:
        return ""

    if forge == "github.com":
        link = f"{git_url}/commit/{upstream_commit}"
    # GitLab or GitLab instances (e.g. gitlab.gnome.org)
    elif "gitlab" in forge:
        link = f"{git_url}/-/commit/{upstream_commit}"
    elif forge == "pagure.io":
        link = f"{git_url}/c/{upstream_commit}"

    return link



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Update selected component from upstream in Fedora
"""

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.config.aliases import get_branches
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)

logger = logging.getLogger(__name__)


@click.command("sync-from-downstream", context_settings=get_context_settings())
@click.option(
    "--dist-git-branch",
    help="Comma separated list of target branches in dist-git to sync from. "
    "(defaults to repo's default branch)",
)
@click.option(
    "--upstream-branch",
    help="Target branch in upstream to sync to. (defaults to repo's default branch)",
)
@click.option(
    "--no-pr",
    is_flag=True,
    default=False,
    help="Do not create a pull request to upstream repository.",
)
@click.option(
    "--fork/--no-fork",
    is_flag=True,
    default=True,
    help="Push to a fork before creating a pull request.",
)
@click.option(
    "--remote-to-push",
    default=None,
    help=(
        "Name of the remote where packit should push. "
        "If this is not specified, push to a fork if the repo can be forked."
    ),
)
@click.option(
    "--force",
    "-f",
    default=False,
    is_flag=True,
    help="Don't discard changes in the git repo by default, unless this is set.",
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="sync back"),
)
@click.option("-x", "--exclude", help="File to exclude from sync", multiple=True)
@click.argument(
    "path_or_url",
    type=LocalProjectParameter(),
    default=os.path.curdir,
)
@pass_config
@cover_packit_exception
@iterate_packages
def sync_from_downstream(
    config,
    dist_git_branch,
    upstream_branch,
    no_pr,
    path_or_url,
    fork,
    remote_to_push,
    exclude,
    force,
    package_config,
):
    """
    Copy synced files from Fedora dist-git into upstream by opening a pull request.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )
    default_dg_branch = api.dg.local_project.git_project.default_branch
    dist_git_branch = dist_git_branch or default_dg_branch
    branches_to_sync = get_branches(
        *dist_git_branch.split(","),
        default_dg_branch=default_dg_branch,
    )
    click.echo(f"Syncing from the following branches: {', '.join(branches_to_sync)}")

    for branch in branches_to_sync:
        api.sync_from_downstream(
            dist_git_branch=branch,
            upstream_branch=upstream_branch,
            no_pr=no_pr,
            fork=fork,
            remote_name=remote_to_push,
            exclude_files=exclude,
            force=force,
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from typing import Optional


class TestCommandConfig:
    """Configuration of test command."""

    def __init__(
        self,
        default_labels: Optional[list[str]] = None,
        default_identifier: Optional[str] = None,
    ):
        self.default_labels = default_labels
        self.default_identifier = default_identifier



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime

import pytest
from flexmock import flexmock

from packit.utils.koji_helper import KojiHelper


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_builds(error):
    nvrs = [f"test-1.{n}-1.fc37" for n in range(3)]

    def getPackageID(*_, **__):
        return 12345

    def listBuilds(*_, **__):
        if error:
            raise Exception
        return [{"nvr": nvr} for nvr in nvrs]

    session = flexmock(getPackageID=getPackageID, listBuilds=listBuilds)
    result = KojiHelper(session).get_nvrs("test", datetime.datetime(2022, 6, 1))
    if error:
        assert result == []
    else:
        assert result == nvrs


@pytest.mark.parametrize(
    "include_candidate, nvr",
    [
        (False, "test-1.0-2.fc40"),
        (True, "test-2.0-1.fc40"),
    ],
)
def test_get_latest_stable_nvr(include_candidate, nvr):
    candidate_tags = {"f40": "f40-updates-candidate"}
    stable_tags = {"f40-updates-candidate": ["f40-updates", "f40"]}
    builds = {
        "f40-updates-candidate": {"nvr": "test-2.0-1.fc40"},
        "f40-updates": {"nvr": "test-1.0-2.fc40"},
        "f40": {"nvr": "test-1.0-1.fc40"},
    }

    koji_helper = KojiHelper(flexmock())
    flexmock(
        koji_helper,
        get_candidate_tag=lambda b: candidate_tags[b],
        get_stable_tags=lambda t: stable_tags[t],
        get_latest_build_in_tag=lambda _, t: builds[t],
    )
    assert koji_helper.get_latest_stable_nvr("test", "f40", include_candidate) == nvr


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_latest_nvr_in_tag(error):
    nvr = "test-1.0-1.fc37"

    def listTagged(*_, **__):
        if error:
            raise Exception
        return [{"nvr": nvr}]

    session = flexmock(listTagged=listTagged)
    result = KojiHelper(session).get_latest_nvr_in_tag("test", "f37-updates-candidate")
    if error:
        assert result is None
    else:
        assert result == nvr


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_build_tags(error):
    tags = ["f37-updates-testing"]

    def listTags(*_, **__):
        if error:
            raise Exception
        return [{"name": t} for t in tags]

    session = flexmock(listTags=listTags)
    result = KojiHelper(session).get_build_tags("test-1.0-1.fc37")
    if error:
        assert result == []
    else:
        assert result == tags


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_build_changelog(error):
    changelog = [
        (1655726400, "Nikola Forró <nforro@redhat.com> - 0.2-1.fc37", "- third entry"),
        (1652702400, "Nikola Forró <nforro@redhat.com> - 0.1-2.fc37", "- second entry"),
        (1648728000, "Nikola Forró <nforro@redhat.com> - 0.1-1.fc37", "- first entry"),
    ]

    def getRPMHeaders(*_, **__):
        if error:
            raise Exception
        result = list(zip(*changelog))
        return {
            "changelogtime": list(result[0]),
            "changelogname": list(result[1]),
            "changelogtext": list(result[2]),
        }

    session = flexmock(getRPMHeaders=getRPMHeaders)
    result = KojiHelper(session).get_build_changelog("test-0.2-1.fc37")
    if error:
        assert result == []
    else:
        assert result == changelog


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_builds_in_tag(error):
    builds = [
        {
            "package_name": "python-specfile",
            "name": "python3-specfile",
            "version": "0.28.0",
            "release": "1.fc39",
        },
    ]

    def listTagged(*_, **__):
        if error:
            raise Exception
        return builds

    session = flexmock(listTagged=listTagged)
    result = KojiHelper(session).get_builds_in_tag("f39-build-side-12345")
    if error:
        assert result == []
    else:
        assert result == builds


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_build_info(error):
    info = {"id": 123456, "name": "test", "nvr": "test-1.0-1.fc39"}

    def getBuild(*_, **__):
        if error:
            raise Exception
        return info

    session = flexmock(getBuild=getBuild)
    for build in [123456, "test-1.0-1.fc39"]:
        result = KojiHelper(session).get_build_info(build)
        if error:
            assert result is None
        else:
            assert result == info


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_tag_info(error):
    info = {"name": "f39-build-side-12345", "id": 12345}

    def getBuildConfig(*_, **__):
        if error:
            raise Exception
        return info

    session = flexmock(getBuildConfig=getBuildConfig)
    result = KojiHelper(session).get_tag_info("f39-build-side-12345")
    if error:
        assert result is None
    else:
        assert result == info


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_create_sidetag(error):
    info = {"name": "f39-build-side-12345", "id": 12345}

    def createSideTag(*_, **__):
        if error:
            raise Exception
        return info

    session = flexmock(
        logged_in=True,
        createSideTag=createSideTag,
    )
    koji_helper = KojiHelper(session)
    flexmock(koji_helper, get_build_target=lambda _: {"build_tag_name": "f39-build"})
    result = koji_helper.create_sidetag("f39")
    if error:
        assert result is None
    else:
        assert result == info


@pytest.mark.parametrize(
    "logged_in",
    [False, True],
)
def test_remove_sidetag(logged_in):
    session = flexmock(logged_in=logged_in)
    session.should_receive("gssapi_login").and_return().times(
        0 if logged_in else 1,
    ).mock()
    session.should_receive("removeSideTag").once()
    KojiHelper(session).remove_sidetag("f39-build-side-12345")


@pytest.mark.parametrize(
    "logged_in",
    [False, True],
)
def test_tag_build(logged_in):
    session = flexmock(logged_in=logged_in)
    session.should_receive("gssapi_login").and_return().times(
        0 if logged_in else 1,
    ).mock()
    session.should_receive("tagBuild").once()
    KojiHelper(session).tag_build("test-1.0-1.fc39", "f39-build-side-12345")


@pytest.mark.parametrize(
    "logged_in",
    [False, True],
)
def test_untag_build(logged_in):
    session = flexmock(logged_in=logged_in)
    session.should_receive("gssapi_login").and_return().times(
        0 if logged_in else 1,
    ).mock()
    session.should_receive("untagBuild").once()
    KojiHelper(session).untag_build("test-1.0-1.fc39", "f39-build-side-12345")


@pytest.mark.parametrize(
    "error",
    [False, True],
)
def test_get_build_target(error):
    target = {
        "id": 123456,
        "build_tag_name": "f39-build",
        "dest_tag_name": "f39-updates-candidate",
    }

    def getBuildTarget(*_, **__):
        if error:
            raise Exception
        return target

    session = flexmock(getBuildTarget=getBuildTarget)
    result = KojiHelper(session).get_build_target("f39")
    if error:
        assert result is None
    else:
        assert result == target


@pytest.mark.parametrize(
    "branch, tag",
    [
        ("f39", "f39-updates-candidate"),
        ("epel9", "epel9-testing-candidate"),
        ("eln", "eln-updates-candidate"),
        ("rawhide", "f41-updates-candidate"),
    ],
)
def test_get_candidate_tag(branch, tag):
    targets = {
        "f39": {
            "build_tag_name": "f39-build",
            "dest_tag_name": "f39-updates-candidate",
        },
        "epel9": {
            "build_tag_name": "epel9-build",
            "dest_tag_name": "epel9-testing-candidate",
        },
        "eln": {
            "build_tag_name": "eln-build",
            "dest_tag_name": "eln-updates-candidate",
        },
        "rawhide": {
            "build_tag_name": "f41-build",
            "dest_tag_name": "f41-updates-candidate",
        },
    }

    def get_build_target(branch, *_, **__):
        return targets[branch]

    koji_helper = KojiHelper()
    flexmock(koji_helper, get_build_target=get_build_target)
    assert koji_helper.get_candidate_tag(branch) == tag


@pytest.mark.parametrize(
    "tag, stable_tags",
    [
        ("f37-updates-candidate", ["f37-updates", "f37"]),
        ("f37-updates-testing", ["f37-updates", "f37"]),
        ("epel8-testing-candidate", ["epel8"]),
        ("epel8-testing", ["epel8"]),
        ("f37-updates", ["f37-updates", "f37"]),
        ("epel8", ["epel8"]),
        ("eln", ["eln", "f41"]),
        ("f40-build-side-12345", ["f40-updates", "f40"]),
    ],
)
def test_get_stable_tags(tag, stable_tags):
    ancestors = {
        "f37-updates-candidate": [{"name": "f37-updates"}, {"name": "f37"}],
        "f37-updates-testing": [{"name": "f37-updates"}, {"name": "f37"}],
        "epel8-testing-candidate": [{"name": "epel8"}],
        "epel8-testing": [{"name": "epel8"}],
        "f37-updates": [{"name": "f37"}],
        "epel8": [],
        "eln": [{"name": "f41"}],
        "f40-build-side-12345": [
            {"name": "f40-build"},
            {"name": "f40-override"},
            {"name": "f40-updates"},
            {"name": "f40"},
        ],
    }

    def getFullInheritance(tag, *_, **__):
        return ancestors[tag]

    session = flexmock(getFullInheritance=getFullInheritance)
    assert KojiHelper(session).get_stable_tags(tag) == stable_tags


@pytest.mark.parametrize(
    "branch, target",
    [
        ("f39", "f39-candidate"),
        ("epel9", "epel9-candidate"),
        ("eln", "eln-candidate"),
        ("rawhide", "rawhide"),
        ("main", "rawhide"),
    ],
)
def test_get_build_target_name(branch, target):
    assert KojiHelper.get_build_target_name(branch) == target


@pytest.mark.parametrize(
    "since, formatted_changelog",
    [
        (
            1652702400,
            "* Mon Jun 20 2022 Nikola Forró <nforro@redhat.com> - 0.2-1\n"
            "- third entry\n",
        ),
        (
            1648728000,
            "* Mon Jun 20 2022 Nikola Forró <nforro@redhat.com> - 0.2-1\n"
            "- third entry\n"
            "\n"
            "* Mon May 16 2022 Nikola Forró <nforro@redhat.com> - 0.1-2\n"
            "- second entry\n",
        ),
        (
            0,
            "* Mon Jun 20 2022 Nikola Forró <nforro@redhat.com> - 0.2-1\n"
            "- third entry\n"
            "\n"
            "* Mon May 16 2022 Nikola Forró <nforro@redhat.com> - 0.1-2\n"
            "- second entry\n"
            "\n"
            "* Thu Mar 31 2022 Nikola Forró <nforro@redhat.com> - 0.1-1\n"
            "- first entry\n",
        ),
    ],
)
def test_format_changelog(since, formatted_changelog):
    changelog = [
        (1655726400, "Nikola Forró <nforro@redhat.com> - 0.2-1", "- third entry"),
        (1652702400, "Nikola Forró <nforro@redhat.com> - 0.1-2", "- second entry"),
        (1648728000, "Nikola Forró <nforro@redhat.com> - 0.1-1", "- first entry"),
    ]
    assert KojiHelper.format_changelog(changelog, since) == formatted_changelog



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import json
import logging
from pathlib import Path
from typing import Callable, Optional, Union

from ogr.abstract import GitProject
from ogr.exceptions import (
    APIException,
    GithubAppNotInstalledError,
)
from yaml import YAMLError, safe_load

from packit.config.common_package_config import CommonPackageConfig, MultiplePackages
from packit.config.job_config import (
    JobConfig,
    JobConfigView,
    JobType,
)
from packit.constants import CONFIG_FILE_NAMES
from packit.exceptions import PackitConfigException

logger = logging.getLogger(__name__)


class PackageConfig(MultiplePackages):
    """
    Config class capturing the configuration of a upstream or
    downstream repository, aka "packit.yaml"

    Attributes:
        jobs: List of job configs.
    """

    def __init__(
        self,
        packages: dict[str, CommonPackageConfig],
        jobs: Optional[list[JobConfig]] = None,
    ):
        self._job_views: list[Union[JobConfig, JobConfigView]] = []
        self._package_config_views: dict[str, "PackageConfigView"] = {}
        self._raw_dict_with_defaults: Optional[dict] = None
        super().__init__(packages)
        # Directly manipulating __dict__ is not recommended.
        # It is done here to avoid triggering __setattr__ and
        # should be removed once support for a single package is
        # dropped from config objects.
        self.__dict__["jobs"] = jobs or []

    def __repr__(self):
        # required to avoid cyclical imports
        from packit.schema import PackageConfigSchema

        s = PackageConfigSchema()
        # For __repr__() return a JSON-encoded string, by using dumps().
        # Mind the 's'!
        return f"PackageConfig: {s.dumps(self)}"

    @classmethod
    def set_defaults(
        cls,
        raw_dict: dict,
        repo_name: Optional[str] = None,
        search_specfile: Optional[Callable[..., Optional[str]]] = None,
        **specfile_search_args,
    ) -> None:
        if not raw_dict.get("specfile_path"):
            default_specfile_path = None
            # we default to <downstream_package_name>.spec
            # https://packit.dev/docs/configuration/#specfile_path
            if downstream_package_name := raw_dict.get("downstream_package_name", None):
                default_specfile_path = f"{downstream_package_name}.spec"
            elif search_specfile:
                default_specfile_path = search_specfile(**specfile_search_args)
            raw_dict["specfile_path"] = default_specfile_path

        # Do this _after_ setting the default for 'specfile_path',
        # so that 'downstream_package_name' is considered for setting
        # 'specfile_path' _only_ when set in packit.yaml.
        if repo_name:
            raw_dict.setdefault("upstream_package_name", repo_name)
            raw_dict.setdefault("downstream_package_name", repo_name)

    @classmethod
    def get_from_dict(
        cls,
        raw_dict: dict,
        config_file_path: Optional[str] = None,
        repo_name: Optional[str] = None,
        search_specfile: Optional[Callable[..., Optional[str]]] = None,
        **specfile_search_args,
    ) -> "PackageConfig":
        # required to avoid cyclical imports

        # we need to process defaults first so they get propagated to JobConfigs
        raw_dict.setdefault("config_file_path", config_file_path)

        if packages := raw_dict.get("packages", None):
            for package in packages.values():
                up_url_key = "upstream_project_url"
                if up_url_key in raw_dict:
                    # propagate it to any monorepo sub-package
                    package[up_url_key] = raw_dict[up_url_key]
                cls.set_defaults(
                    package,
                    repo_name,
                    search_specfile,
                    **specfile_search_args,
                )
        else:
            cls.set_defaults(
                raw_dict,
                repo_name,
                search_specfile,
                **specfile_search_args,
            )

        return cls.get_from_dict_without_setting_defaults(raw_dict)

    @classmethod
    def get_from_dict_without_setting_defaults(
        cls,
        raw_dict: dict,
    ) -> "PackageConfig":
        # required to avoid cyclical imports
        from packit.schema import PackageConfigSchema

        package_config = PackageConfigSchema().load(raw_dict)
        package_config._raw_dict_with_defaults = raw_dict
        return cls.post_load(package_config)

    def get_raw_dict_with_defaults(self):
        return self._raw_dict_with_defaults

    @classmethod
    def post_load(cls, package_config: "PackageConfig") -> "PackageConfig":
        if not package_config:
            return None

        package_config_views: dict[str, "PackageConfigView"] = {}
        for name, package in package_config.packages.items():
            # filter out job data for the package
            jobs = [
                job for job in package_config.get_job_views() if name in job.packages
            ]
            package_config_views[name] = PackageConfigView(
                packages={name: package},
                jobs=jobs,
            )
        package_config.set_package_config_views(package_config_views)

        return package_config

    def get_copr_build_project_value(self) -> Optional[str]:
        """
        get copr project name from this first copr job
        this is only used when invoking copr builds from CLI
        """
        projects_list = [
            job.project
            for job in self.jobs
            if job.type == JobType.copr_build and job.project
        ]
        if not projects_list:
            return None

        if len(set(projects_list)) > 1:
            logger.warning(
                f"You have defined multiple copr projects to build in, we are going "
                f"to pick the first one: {projects_list[0]}, reorder the job definitions"
                f" if this is not the one you want.",
            )
        return projects_list[0]

    def get_propose_downstream_dg_branches_value(
        self,
        pull_from_upstream=False,
    ) -> Optional[list]:
        type = (
            JobType.pull_from_upstream
            if pull_from_upstream
            else JobType.propose_downstream
        )
        for job in self.jobs:
            if job.type == type:
                return job.dist_git_branches
        return []

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        # required to avoid cyclical imports
        from packit.schema import PackageConfigSchema

        s = PackageConfigSchema()
        # Compare the serialized objects.
        serialized_self = s.dump(self)
        serialized_other = s.dump(other)
        logger.debug(f"our configuration:\n{serialized_self}")
        logger.debug(f"the other configuration:\n{serialized_other}")
        return serialized_self == serialized_other

    def get_job_views(self) -> list[Union[JobConfig, JobConfigView]]:
        """Get jobs views on a single package.
        If a JobConfig reference more than a package, then
        split it in many JobConfigView(s) one for any package
        """
        if self._job_views:
            return self._job_views

        for job in self.jobs:
            if len(self.packages) > 1:
                for name in job.packages:
                    job_view = JobConfigView(job, name)
                    self._job_views.append(job_view)
            else:
                self._job_views.append(job)
        return self._job_views

    def get_package_config_views(self) -> dict[str, "PackageConfigView"]:
        """Return a dictionary of package name -> PackageConfigView
        every PackageConfigView holds just one package (the named one)
        and its associated jobs.
        A Monorepo PackageConfig will be splitted in many of them.

        NOTE: not using a property because of the custom __getattr__
        """
        return self._package_config_views

    def set_package_config_views(self, value: dict[str, "PackageConfigView"]):
        """Set a dictionary of package name -> PackageConfigView
        every PackageConfigView holds just one package (the named one)
        and its associated jobs.
        A Monorepo PackageConfig will be split in many of them.

        NOTE: not using a property because of the custom __setattr__
        """
        self._package_config_views = value

    def get_package_config_for(
        self,
        job_config: Union[JobConfigView, JobConfig],
    ) -> Union["PackageConfigView", "PackageConfig"]:
        """Select the PackageConfigView for the given JobConfig in
        a multiple packages config.
        """
        package_config_views = self.get_package_config_views()
        if not package_config_views or not job_config:
            # the package config views were not initialized
            # we can continue if this is not a monorepo
            if len(self.packages) == 1:
                return self
        else:
            if not isinstance(job_config, JobConfigView):
                # the job config should have just one package,
                # choose that one
                return package_config_views[next(iter(job_config.packages.keys()))]

        return package_config_views[job_config.package]


class PackageConfigView(PackageConfig):
    """A PackageConfig which holds:
    - one single package config (no more than one CommonPackageConfig object)
    - only jobs related with the given package config

    The PackageConfig is responsible to build this object
    filtering out all the data related to a single package
    """

    def __init__(
        self,
        packages: dict[str, CommonPackageConfig],
        jobs: Optional[list[JobConfig]] = None,
    ):
        if len(packages) > 1:
            logger.error(
                "The PackageConfigView class deals with just one single package",
            )
        super().__init__(packages, jobs)


def find_packit_yaml(
    *directory: Union[Path, str],
    try_local_dir_first: bool = False,
    try_local_dir_last: bool = False,
) -> Path:
    """
    Find packit config in provided directories.

    Args:
        directory: List of directories where we should look for the config.
        try_local_dir_first: If set to `True` check the current working directory
            first.

            Defaults to `False`.
        try_local_dir_last: If set to `True` check the current working directory
            last.

            Defaults to `False`.

    Returns:
        Path to the config.

    Raises:
        PackitConfigException: If no config is found.
    """
    directories = [Path(config_dir) for config_dir in directory]
    cwd = Path.cwd()

    if try_local_dir_first and try_local_dir_last:
        logger.error(
            "Ambiguous usage of 'try_local_dir_first' and 'try_local_dir_last'.",
        )

    if try_local_dir_first:
        if cwd in directories:
            directories.remove(cwd)
        directories.insert(0, cwd)

    if try_local_dir_last:
        if cwd in directories:
            directories.remove(cwd)
        directories.append(cwd)

    for config_dir in directories:
        for config_file_name in CONFIG_FILE_NAMES:
            config_file_name_full = config_dir / config_file_name
            if config_file_name_full.is_file():
                logger.debug(f"Local package config found: {config_file_name_full}")
                return config_file_name_full
    raise PackitConfigException("No packit config found.")


def load_packit_yaml(
    config_file_path: Optional[Path] = None,
    raw_text: str = "",
) -> dict:
    """
    Use yaml.safe_load to parse provided text as yaml

    When config_file_path is set, read the file, otherwise process content of raw_text.

    Args:
        config_file_path: local path to .packit.yaml
        raw_text: content of .packit.yaml

    Raises:
        PackitConfigException when something goes wrong while loading the config

    Returns:
        Dict with the file content
    """
    if config_file_path:
        raw_text = config_file_path.read_text()
    try:
        # safe_load() returns None when the file is empty, but this needs
        # to return a dict.
        return safe_load(raw_text) or {}
    except YAMLError as ex:
        logger.error(f"Cannot load package config {config_file_path}.")
        if hasattr(ex, "problem_mark"):
            msg = (
                "  parser says\n"
                + str(ex.problem_mark)
                + "\n  "
                + str(ex.problem)  # type: ignore
            )  # type: ignore
            if ex.context is not None:  # type: ignore
                msg += f" {ex.context!s}"  # type: ignore
            logger.error(msg)
        else:
            logger.error(f"parser says: {ex!r}.")
        raise PackitConfigException("Please correct data and retry.") from ex


def get_local_package_config(
    *directory: Union[Path, str],
    repo_name: Optional[str] = None,
    try_local_dir_first: bool = False,
    try_local_dir_last: bool = False,
    package_config_path: Optional[Union[Path, str]] = None,
) -> PackageConfig:
    """
    Finds and loads packit config from the provided directories.

    Args:
        directory: List of directories to look for packit config in.
        repo_name: Name of the git repository (default for project name).
        try_local_dir_first: If set to `True` check the current working directory
            first.

            Defaults to `False`.
        try_local_dir_last: If set to `True` check the current working directory
            last.

            Defautls to `False`.
        package_config_path: Path to the package configuration file.

    Returns:
        Loaded config as a `PackageConfig`.
    """

    if package_config_path:
        config_file_name = Path(package_config_path)
    else:
        config_file_name = find_packit_yaml(
            *directory,
            try_local_dir_first=try_local_dir_first,
            try_local_dir_last=try_local_dir_last,
        )

    loaded_config = load_packit_yaml(config_file_path=config_file_name)

    return parse_loaded_config(
        loaded_config=loaded_config,
        config_file_path=config_file_name.name,
        repo_name=repo_name,
        search_specfile=get_local_specfile_path,
        dir=config_file_name.parent,
    )


def find_remote_package_config(
    project: GitProject,
    ref: Optional[str] = None,
) -> Optional[str]:
    """
    Check if there's a package config file
    in the given project (top-level directory only).

    Args:
        project: ogr Git-project object.
        ref: Optional ref at which the config should be searched for.

    Returns:
        Name of the found config file or None if there's no such file.
    """

    try:
        candidates = set(project.get_files(ref=ref, recursive=False))
    except GithubAppNotInstalledError:
        logger.warning(
            "The Packit GitHub App is not installed"
            f"for the {project.full_repo_name!r} repository.",
        )
        return None
    except APIException as ex:
        if ex.response_code == 404:
            # we couldn't find the project or git reference
            logger.warning(f"No project or ref was found: {ex}")
            return None
        raise

    try:
        package_config_name = (candidates & CONFIG_FILE_NAMES).pop()
    except KeyError:
        logger.warning(
            f"No config file ({CONFIG_FILE_NAMES}) found on ref {ref!r} "
            f"of the {project.full_repo_name!r} repository.",
        )
        return None

    logger.debug(
        f"Found a config file {package_config_name!r} "
        f"on ref {ref!r} "
        f"of the {project.full_repo_name!r} repository.",
    )
    return package_config_name


def get_package_config_from_repo(
    project: GitProject,
    ref: Optional[str] = None,
    package_config_path: Optional[str] = None,
) -> Optional[PackageConfig]:
    """Search for the package config in a remote repo, load it and return
    the package configuration object.

    Args:
        project: ogr Git-project object.
        ref: Optional ref at which the config should be searched for.
        package_config_path: path of the package config, relative to the repo root.
            Load and parse this when specified instead of searching for one.

    Returns:
        PackageConfig object constructed from the config file found in
        the repo or None if there's no package config in the repo.
    """
    if not (
        package_config_path := package_config_path
        or find_remote_package_config(project, ref)
    ):
        return None

    try:
        config_file_content = project.get_file_content(
            path=package_config_path,
            ref=ref,
        )
    except FileNotFoundError:
        logger.warning(
            f"No config file {package_config_path!r} found on ref {ref!r} "
            f"of the {project.full_repo_name!r} repository.",
        )
        return None
    loaded_config = load_packit_yaml(raw_text=config_file_content)

    return parse_loaded_config(
        loaded_config=loaded_config,
        config_file_path=package_config_path,
        repo_name=project.repo,
        search_specfile=get_specfile_path_from_repo,
        project=project,
        ref=ref,
    )


def parse_loaded_config(
    loaded_config: dict,
    config_file_path: Optional[str] = None,
    repo_name: Optional[str] = None,
    spec_file_path: Optional[str] = None,
    search_specfile: Optional[Callable[..., Optional[str]]] = None,
    **specfile_search_args,
) -> PackageConfig:
    """Tries to parse the config to PackageConfig."""
    logger.debug(
        f"Package config before loading:\n{json.dumps(loaded_config, indent=4)}",
    )

    try:
        return PackageConfig.get_from_dict(
            raw_dict=loaded_config,
            config_file_path=config_file_path,
            repo_name=repo_name,
            search_specfile=search_specfile,
            **specfile_search_args,
        )
    except Exception as ex:
        logger.error(f"Cannot parse package config. {ex}.")
        raise PackitConfigException(f"Cannot parse package config. {ex!r}") from ex


def get_local_specfile_path(
    dir: Path,
    exclude: Optional[list[str]] = None,
) -> Optional[str]:
    """
    Get the path to the local specfile if present. If specfile is not found in
    the directory itself, search for it recursively (rglob).

    Args:
        dir: Directory to find the specfile in.
        exclude: Directories to be excluded from the recursive search.

            Defaults to `["tests"]`.

    Returns:
        Path (relative to the `dir`) of the specfile that was found first.
    """
    files = [path.relative_to(dir) for path in dir.glob("*.spec")] or [
        path.relative_to(dir) for path in dir.rglob("*.spec")
    ]

    # Don't take files found in exclude
    sexclude = set(exclude) if exclude else {"tests"}
    files = [f for f in files if f.parts[0] not in sexclude]

    if len(files) > 0:
        logger.debug(f"Local spec files found: {files}. Taking: {files[0]}")
        return str(files[0])

    return None


def get_specfile_path_from_repo(
    project: GitProject,
    ref: Optional[str] = None,
) -> Optional[str]:
    """
    Get the path of the specfile in the given repo if present.

    Args:
        project: Repository to look for the specfile in.
        ref: Git reference where the specfile is to be found.

            Defaults to repository's default branch.

    Returns:
        Path to the specfile or `None` if not found.
    """
    spec_files = project.get_files(ref=ref, filter_regex=r".+\.spec$")

    if not spec_files:
        logger.debug(f"No spec file found in {project.full_repo_name!r}")
        return None
    return spec_files[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.patches import PatchMetadata, commit_message, remove_prefixes


@pytest.fixture
def patch(tmp_path):
    patch_file = tmp_path / "patch.patch"
    patch_file.write_text(
        """\
From 477fb1b17ee5fa84913102e964239c0d28019b8a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Everyday=20Programmer?= <eprog@redhat.com>
Date: Wed, 18 Aug 2021 10:06:09 +0200
Subject: [PATCH 1/2] Add some content
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

This is an explanation why the content was added.

Signed-off-by: Everyday Programmer <eprog@redhat.com>
---
 a.file | 1 +
 b.file | 1 +
 2 files changed, 2 insertions(+)

diff --git a/a.file b/a.file
index e69de29..8a1853f 100644
--- a/a.file
+++ b/a.file
@@ -0,0 +1 @@
+Some content in file A.
diff --git a/b.file b/b.file
index e69de29..e892e75 100644
--- a/b.file
+++ b/b.file
@@ -0,0 +1 @@
+Some content in file B.
--
2.31.1
""",
    )
    return patch_file


@pytest.fixture
def patch_without_body(tmp_path):
    patch_file = tmp_path / "without_body.patch"
    patch_file.write_text(
        """\
From 477fb1b17ee5fa84913102e964239c0d28019b8a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Everyday=20Programmer?= <eprog@redhat.com>
Date: Wed, 18 Aug 2021 10:06:09 +0200
Subject: [PATCH 1/2] Add some content
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

---
 a.file | 1 +
 b.file | 1 +
 2 files changed, 2 insertions(+)

diff --git a/a.file b/a.file
index e69de29..8a1853f 100644
--- a/a.file
+++ b/a.file
@@ -0,0 +1 @@
+Some content in file A.
diff --git a/b.file b/b.file
index e69de29..e892e75 100644
--- a/b.file
+++ b/b.file
@@ -0,0 +1 @@
+Some content in file B.
--
2.31.1
""",
    )
    return patch_file


@pytest.fixture
def patch_with_bytes(tmp_path):
    patch_file = tmp_path / "with_bytes.patch"
    patch_file.write_bytes(
        b"""\
From 477fb1b17ee5fa84913102e964239c0d28019b8a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Everyday=20Programmer?= <eprog@redhat.com>
Date: Wed, 18 Aug 2021 10:06:09 +0200
Subject: [PATCH 1/2] Add some content
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

This is an explanation why the content was added.

Signed-off-by: Everyday Programmer <eprog@redhat.com>
---
 a.file | 1 +
 b.file | 1 +
 2 files changed, 2 insertions(+)

diff --git a/a.file b/a.file
index e69de29..8a1853f 100644
--- a/a.file
+++ b/a.file
@@ -0,0 +1 @@
+Some content \xc0 in file A.
diff --git a/b.file b/b.file
index e69de29..e892e75 100644
--- a/b.file
+++ b/b.file
@@ -0,0 +1 @@
+Some content in file B.
--
2.31.1
""",
    )
    return patch_file


@pytest.fixture
def patch_with_meta(tmp_path):
    patch_file = tmp_path / "with_meta.patch"
    patch_file.write_text(
        """\
From 477fb1b17ee5fa84913102e964239c0d28019b8a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Everyday=20Programmer?= <eprog@redhat.com>
Date: Wed, 18 Aug 2021 10:06:09 +0200
Subject: [PATCH 1/2] Add some content
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

This is an explanation why the content was added.

Patch-name: clev:er.patch
Patch-id: 200
Patch-status: |
    # The: empty line
    #
    # needs to be here
Patch-present-in-specfile: true
Ignore-patch: true
No-prefix: true
Signed-off-by: Everyday Programmer <eprog@redhat.com>
---
 a.file | 1 +
 b.file | 1 +
 2 files changed, 2 insertions(+)

diff --git a/a.file b/a.file
index e69de29..8a1853f 100644
--- a/a.file
+++ b/a.file
@@ -0,0 +1 @@
+Some content in file A.
diff --git a/b.file b/b.file
index e69de29..e892e75 100644
--- a/b.file
+++ b/b.file
@@ -0,0 +1 @@
+Some content in file B.
--
2.31.1
""",
    )
    return patch_file


@pytest.fixture
def commit_message_file(tmp_path):
    file = tmp_path / "commit_message.file"
    file.write_text(
        """\
[PATCH 1/2] Add some content

This is an explanation why the content was added.

Signed-off-by: Everyday Programmer <eprog@redhat.com>
""",
    )
    return file


def test_remove_prefixes(patch):
    remove_prefixes(patch)
    assert (
        patch.read_text()
        == """\
From 477fb1b17ee5fa84913102e964239c0d28019b8a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Everyday=20Programmer?= <eprog@redhat.com>
Date: Wed, 18 Aug 2021 10:06:09 +0200
Subject: [PATCH 1/2] Add some content
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

This is an explanation why the content was added.

Signed-off-by: Everyday Programmer <eprog@redhat.com>
---
 a.file | 1 +
 b.file | 1 +
 2 files changed, 2 insertions(+)

diff --git a.file a.file
index e69de29..8a1853f 100644
--- a.file
+++ a.file
@@ -0,0 +1 @@
+Some content in file A.
diff --git b.file b.file
index e69de29..e892e75 100644
--- b.file
+++ b.file
@@ -0,0 +1 @@
+Some content in file B.
--
2.31.1
"""
    )


body = """This is an explanation why the content was added."""

body_with_trailer = f"""{body}

Signed-off-by: Everyday Programmer <eprog@redhat.com>"""


@pytest.mark.parametrize(
    "patch_file, strip_subject_prefix, strip_trailers, body",
    [
        ("patch", None, "Signed-off-by: Everyday Programmer <eprog@redhat.com>", body),
        ("patch", "PATCH", None, body_with_trailer),
        (
            "patch_with_bytes",
            None,
            "Signed-off-by: Everyday Programmer <eprog@redhat.com>",
            body,
        ),
        ("patch_with_bytes", "PATCH", None, body_with_trailer),
        (
            "patch_without_body",
            "PATCH",
            "Signed-off-by: Everyday Programmer <eprog@redhat.com>",
            "",
        ),
        (
            "commit_message_file",
            None,
            "Signed-off-by: Everyday Programmer <eprog@redhat.com>",
            body,
        ),
        ("commit_message_file", "PATCH", None, body_with_trailer),
    ],
)
def test_commit_message(
    patch_file,
    strip_subject_prefix,
    strip_trailers,
    body,
    request,
):
    file = request.getfixturevalue(patch_file)
    subject = f"{'' if strip_subject_prefix else '[PATCH 1/2] '}Add some content"
    empty_line = "\n\n"
    assert (
        commit_message(
            file,
            strip_subject_prefix=strip_subject_prefix,
            strip_trailers=strip_trailers,
        )
        == f"{subject}{empty_line + body if body else ''}"
    )


@pytest.mark.parametrize(
    "patch_file, meta_fields",
    [
        (
            "patch",
            {
                "name": "patch.patch",
                "description": "Add some content\n\n"
                "This is an explanation why the content was added.",
            },
        ),
        (
            "patch_with_meta",
            {
                "name": "clev:er.patch",
                "description": "# The: empty line\n#\n# needs to be here",
                "present_in_specfile": True,
                "ignore": True,
                "patch_id": 200,
                "no_prefix": True,
                "metadata_defined": True,
            },
        ),
    ],
)
def test_from_patch(patch_file, meta_fields, request):
    patch_path = request.getfixturevalue(patch_file)
    assert PatchMetadata.from_patch(str(patch_path)) == PatchMetadata(
        path=patch_path,
        **meta_fields,
    )


def test_from_git_trailers():
    commit = flexmock(
        message="""\
Add a test commit

Patch-name: test.patch
Signed-off-by: Everyday Programmer <eprog@redhat.com>
""",
    )
    patch_meta = PatchMetadata.from_git_trailers(commit)
    assert patch_meta.name == "test.patch"
